from collections.abc import Mapping
from itertools import combinations

from torch import BoolTensor, Tensor, cat, nn

from salt.models.attention import MultiheadAttention
from salt.models.dense import Dense
from salt.stypes import Tensors


class TransformerEncoderLayer(nn.Module):
    """A transformer encoder layer.

    We switched from normformer style to pre-ln style in MR !132.

    It contains:
    - Multihead attention block
    - A feedforward network (which can take optional context information)
    - Layernorms
    - Residual connections
    - Optional edge update blocks
    """

    def __init__(
        self,
        embed_dim: int,
        mha_config: Mapping,
        dense_config: Mapping | None = None,
        context_dim: int = 0,
        edge_embed_dim: int = 0,
        update_edges: bool = False,
        muP: bool = False,
    ) -> None:
        """Init method of TransformerEncoderBlock.

        Parameters
        ----------
        embed_dim : int
            The embedding dimension of the transformer block
        mha_config : int
            Configuration for the MultiheadAttention block
        dense_config : Mapping
            Configuration for the Dense network
        context_dim: int
            The size of the context tensor
        edge_embed_dim : int
            The embedding dimension of the transformer block for edge features
        update_edges: bool
            Value indicating whether to update edge features via attention
        muP: bool, optional,
            Whether to use the muP parametrisation (impacts initialisation).
        """
        super().__init__()
        self.embed_dim = embed_dim
        self.edge_embed_dim = edge_embed_dim
        self.update_edges = update_edges
        self.muP = muP

        # The main blocks in the transformer
        self.mha = MultiheadAttention(
            embed_dim,
            edge_embed_dim=edge_embed_dim,
            update_edges=update_edges,
            muP=self.muP,
            **mha_config,
        )
        if dense_config:
            self.dense = Dense(
                input_size=embed_dim,
                output_size=embed_dim,
                context_size=context_dim,
                muP=self.muP,
                **dense_config,
            )
        else:
            self.register_buffer("dense", None)

        # The multiple normalisation layers to keep it all stable
        self.norm1 = nn.LayerNorm(embed_dim)
        self.norm2 = nn.LayerNorm(embed_dim)
        if self.edge_embed_dim > 0:
            self.enorm1 = nn.LayerNorm(edge_embed_dim)
            if self.update_edges:
                self.enorm2 = nn.LayerNorm(edge_embed_dim)

    def forward(
        self,
        x: Tensor,
        edge_x: Tensor | None = None,
        pad_mask: BoolTensor | None = None,
        context: Tensor | None = None,
        attn_mask: BoolTensor | None = None,
        attn_bias: Tensor | None = None,
    ) -> Tensor:
        if edge_x is not None:
            xi, edge_xi = self.mha(
                self.norm1(x),
                edges=self.enorm1(edge_x),
                q_mask=pad_mask,
                attn_mask=attn_mask,
                attn_bias=attn_bias,
            )
        else:
            xi = self.mha(
                self.norm1(x),
                q_mask=pad_mask,
                attn_mask=attn_mask,
                attn_bias=attn_bias,
            )

        x = x + xi
        if self.update_edges:
            edge_x = edge_x + self.enorm2(edge_xi)
        if self.dense:
            x = x + self.dense(self.norm2(x), context)

        if edge_x is not None:
            return x, edge_x

        return x


class TransformerEncoder(nn.Module):
    """A stack of N transformer encoder layers followed by a final
    normalisation step.

    A message passing network which does not change the dimension of the
    nodes, and uses multiheaded attention to facilitate message
    importance.
    """

    def __init__(
        self,
        embed_dim: int,
        num_layers: int,
        mha_config: Mapping,
        dense_config: Mapping | None = None,
        context_dim: int = 0,
        out_dim: int = 0,
        edge_embed_dim: int = 0,
        update_edges: bool = False,
        muP: bool = False,
    ) -> None:
        """Transformer encoder module.

        Parameters
        ----------
        embed_dim : int
            Feature size for input, output, and all intermediate layers
        num_layers : int
            Number of encoder layers used
        mha_config : nn.Module
            Keyword arguments for the mha block
        dense_config: Mapping
            Keyword arguments for [`salt.models.Dense`][salt.models.Dense]
        context_dim: int
            Dimension of the context inputs
        out_dim: int
            If set, a final linear layer resizes the outputs of the model
        edge_embed_dim : int
            Feature size for input and output of edge features
        update_edges: bool
            If set, edge features are updated in each encoder layer
        muP: bool, optional,
            Whether to use the muP parametrisation (impacts initialisation).
        """
        super().__init__()
        self.embed_dim = embed_dim
        self.edge_embed_dim = edge_embed_dim
        self.num_layers = num_layers
        self.out_dim = out_dim
        self.update_edges = update_edges
        self.muP = muP
        self.featurewise = nn.ModuleList()

        self.layers = nn.ModuleList([
            TransformerEncoderLayer(
                embed_dim,
                mha_config,
                dense_config,
                context_dim,
                edge_embed_dim,
                update_edges if i != num_layers - 1 else False,
                self.muP,
            )
            for i in range(num_layers)
        ])
        self.final_norm = nn.LayerNorm(embed_dim)

        # For resizing the output tokens
        if self.muP:
            assert self.out_dim, "Need the out_dim layer for muP, \
                as this is the last layer of the muP-part of the model"
        if self.out_dim:
            if self.muP:
                from mup import MuReadout

                self.final_linear = MuReadout(self.embed_dim, self.out_dim)
                self.final_linear.bias.data.zero_()
                self.final_linear.weight.data.zero_()
            else:
                self.final_linear = nn.Linear(self.embed_dim, self.out_dim)

    def forward(
        self,
        x: Tensor | dict,
        edge_x: Tensor = None,
        pad_mask: Tensor | dict | None = None,
        inputs: Tensors = None,
        **kwargs,
    ) -> Tensor:
        """Pass the input through all layers sequentially."""
        if isinstance(x, dict):
            x = cat(list(x.values()), dim=1)

        if isinstance(pad_mask, dict):
            pad_mask = cat(list(pad_mask.values()), dim=1)

        for i, layer in enumerate(self.layers):
            if len(self.featurewise) > 0:
                x = self.featurewise[i](inputs, x)
            if edge_x is not None:
                x, edge_x = layer(x, edge_x, pad_mask=pad_mask, **kwargs)
            else:
                x = layer(x, pad_mask=pad_mask, **kwargs)
        x = self.final_norm(x)

        # optional resizing layer
        if self.out_dim:
            x = self.final_linear(x)
        return x


class TransformerCrossAttentionLayer(TransformerEncoderLayer):
    """A transformer encoder layer with cross-attention."""

    def __init__(
        self,
        embed_dim: int,
        mha_config: Mapping,
        dense_config: Mapping | None = None,
        context_dim: int = 0,
        muP: bool = False,
    ) -> None:
        super().__init__(embed_dim, mha_config, dense_config, context_dim, muP)
        self.norm0 = nn.LayerNorm(embed_dim)

    def forward(  # type: ignore[override]
        self,
        query: Tensor,
        key_value: Tensor,
        query_mask: BoolTensor | None = None,
        key_value_mask: BoolTensor | None = None,
        context: Tensor | None = None,
    ) -> Tensor:
        additive = self.mha(
            self.norm1(query),
            self.norm0(key_value),
            q_mask=query_mask,
            kv_mask=key_value_mask,
        )
        if self.dense:
            additive = self.dense(self.norm2(additive), context)
        # The cross attention does not return residual+additive but only the additive
        # It is ment to be used as Ai+1 += CA.forward(Ai, Bi) if one wants to
        # have a residual connection
        # and Ai+1 = CA.forward(Ai, Bi) if a simple feed forward connection is intended
        # we also recoomend to use clones of Ai and Bi and avoid A += CA.forward(A, B)
        return additive


class TransformerCrossAttentionEncoder(nn.Module):
    """A stack of N transformer encoder layers interspersed with cross-attention
    layers between inputs types.
    """

    def __init__(
        self,
        input_names: list[str],
        embed_dim: int,
        num_layers: int,
        mha_config: Mapping,
        sa_dense_config: Mapping | None = None,
        ca_dense_config: Mapping | None = None,
        context_dim: int = 0,
        out_dim: int = 0,
        ca_every_layer: bool = False,
        update_edges: bool = False,
        muP: bool = False,
    ):
        super().__init__()
        self.input_names = input_names
        self.embed_dim = embed_dim
        self.num_layers = num_layers
        self.out_dim = out_dim
        self.ca_every_layer = ca_every_layer
        self.update_edges = update_edges
        self.muP = muP

        # Layers for each input type
        # need to use ModuleDict so device is set correctly
        self.type_layers = nn.ModuleDict({
            input_name: nn.ModuleList([
                TransformerEncoderLayer(
                    embed_dim,
                    mha_config,
                    sa_dense_config,
                    context_dim,
                    update_edges if i != num_layers - 1 else False,
                    muP,
                )
                for i in range(num_layers)
            ])
            for input_name in self.input_names
        })

        ca_layers = num_layers if ca_every_layer else 2 if num_layers > 1 else 1

        # module dict only supports string keys
        self.cross_layers = nn.ModuleDict({
            f"{input_name1}_{input_name2}": nn.ModuleList([
                TransformerCrossAttentionLayer(
                    embed_dim,
                    mha_config,
                    ca_dense_config,
                    context_dim,
                    muP,
                )
                for _ in range(ca_layers)
            ])
            for input_name1, input_name2 in combinations(self.input_names, 2)
        })

        self.final_norm = nn.LayerNorm(embed_dim)

        if self.muP:
            assert self.out_dim, "Need the out_dim layer for muP, \
                as this is the last layer of the muP-part of the model"

        # For resizing the output tokens
        if self.out_dim:
            if muP:
                from mup import MuReadout

                self.final_linear = MuReadout(self.embed_dim, self.out_dim)
                self.final_linear.bias.data.zero_()
                self.final_linear.weight.data.zero_()
            else:
                self.final_linear = nn.Linear(self.embed_dim, self.out_dim)

    def forward(
        self, x: dict[str, Tensor], pad_mask: dict[str, Tensor], edge_x: Tensor = None, **kwargs
    ) -> Tensor:
        """Pass the input through all layers sequentially."""
        if edge_x is not None:
            raise ValueError("Edge updates of Cross Attention Encoder not yet supported")

        # Merge inputs as specified
        for i in range(self.num_layers):
            # Self-attention for each type
            for it in self.input_names:
                x[it] = self.type_layers[it][i](x[it], pad_mask=pad_mask[it], **kwargs)

            # Cross-attention between pairs of types - symmetric update
            # i % len(self.cross_layers[layer_key]) evals to 0 for first layer
            # and final layer and 1 everywhere else to give behaviour we want
            prev_x = {k: x[k].clone() for k in x}
            if self.ca_every_layer:
                for it1, it2 in combinations(self.input_names, 2):
                    layer_key = f"{it1}_{it2}"
                    x[it1] += self.cross_layers[layer_key][i](
                        prev_x[it1], prev_x[it2], pad_mask[it1], pad_mask[it2], **kwargs
                    )
                    x[it2] += self.cross_layers[layer_key][i](
                        prev_x[it2], prev_x[it1], pad_mask[it2], pad_mask[it1], **kwargs
                    )
            elif i in {0, self.num_layers - 1}:
                for it1, it2 in combinations(self.input_names, 2):
                    layer_key = f"{it1}_{it2}"
                    x[it1] += self.cross_layers[layer_key][i // (self.num_layers - 1)](
                        prev_x[it1], prev_x[it2], pad_mask[it1], pad_mask[it2], **kwargs
                    )
                    x[it2] += self.cross_layers[layer_key][i // (self.num_layers - 1)](
                        prev_x[it2], prev_x[it1], pad_mask[it2], pad_mask[it1], **kwargs
                    )

        # Apply final normalization
        for it in self.input_names:
            x[it] = self.final_norm(x[it])

        # Optional resizing layer
        if self.out_dim:
            for it in self.input_names:
                x[it] = self.final_linear(x[it])

        return x



import sys
from pathlib import Path

import h5py
import pytest
import yaml

from salt.main import main
from salt.onnx.to_onnx import main as to_onnx
from salt.utils.get_onnx_metadata import main as get_onnx_metadata
from salt.utils.inputs import write_dummy_file, write_dummy_norm_dict

w = "ignore::lightning.fabric.utilities.warnings.PossibleUserWarning:"
CONFIG = "GN2.yaml"
TAU_CONFIGS = {"GN2.yaml", "GN3_baseline.yaml"}


def run_train(tmp_path, config_path, train_args, do_xbb=False, do_muP=False, inc_params=False):
    incl_taus = config_path.name in TAU_CONFIGS
    tmp_path = Path(tmp_path)
    train_h5_path = tmp_path / "dummy_train_inputs.h5"
    nd_path = tmp_path / "dummy_norm_dict.yaml"
    cd_path = tmp_path / "dummy_class_dict.yaml"
    write_dummy_norm_dict(nd_path, cd_path)
    write_dummy_file(train_h5_path, nd_path, do_xbb, incl_taus, inc_params)

    args = ["fit"]
    args += [f"--config={config_path}"]
    args += [f"--data.norm_dict={nd_path}"]
    args += [f"--data.class_dict={cd_path}"]
    args += [f"--data.train_file={train_h5_path}"]
    args += [f"--data.val_file={train_h5_path}"]
    args += ["--data.num_train=500"]
    args += ["--data.num_val=200"]
    args += ["--data.batch_size=100"]
    args += ["--data.num_workers=0"]
    args += ["--trainer.max_epochs=1"]
    args += ["--trainer.accelerator=cpu"]
    args += ["--trainer.devices=1"]
    args += [f"--trainer.default_root_dir={tmp_path}"]
    args += ["--trainer.logger.offline=True"]

    if train_args:
        args += train_args

    if do_muP:
        from salt.utils.muP_utils.main_muP import main as main_muP

        # skip fit and callbacks
        main_muP(args=args[1:-2])

    main(args)


def run_eval(tmp_path, train_config_path, nd_path, do_xbb=False):
    test_h5_path = Path(tmp_path) / "dummy_test_sample_inputs.h5"
    write_dummy_file(test_h5_path, nd_path, do_xbb)

    # Modify the output config to force writing tracks in the prediction writer
    with open(train_config_path) as f:
        config = yaml.safe_load(f)
        for callback in config["trainer"]["callbacks"]:
            if "PredictionWriter" in callback["class_path"]:
                callback["init_args"]["write_tracks"] = True
                break
    with open(train_config_path, "w") as f:
        yaml.dump(config, f)
    args = ["test"]

    args += [f"--config={train_config_path}"]
    args += [f"--data.test_file={test_h5_path}"]
    args += ["--data.num_test=1000"]
    args += ["--data.batch_size=100"]
    main(args)

    # check output h5 files are produced
    h5_dir = train_config_path.parent / "ckpts"
    h5_files = [f for f in h5_dir.iterdir() if f.suffix == ".h5"]
    assert len(h5_files) == 1
    h5_file = h5_files[0]
    with h5py.File(h5_file, "r") as f:
        assert "jets" in f
        assert len(f["jets"]) == 1000
        if "GN2" in str(train_config_path):
            assert "tracks" in f
            assert len(f["tracks"]) == 1000

        if "maskformer" in str(train_config_path):
            assert "truth_hadrons" in f
            assert len(f["truth_hadrons"]) == 1000
            assert f["truth_hadrons"].shape[1] == 5
            required_keys = {
                "MaskFormer_pt",
                "MaskFormer_deta",
                "MaskFormer_dphi",
                "MaskFormer_mass",
                "MaskFormer_Lxy",
                "MaskFormer_pb",
                "MaskFormer_pc",
                "MaskFormer_pnull",
                "class_label",
            }
            print(set(required_keys) - set(f["truth_hadrons"].dtype.names))
            print(set(f["truth_hadrons"].dtype.names) - set(required_keys))
            assert all(k in f["truth_hadrons"].dtype.names for k in required_keys)

            assert "object_masks" in f
            assert f["object_masks"].shape == (1000, 5, 40)
            assert "mask_logits" in f["object_masks"].dtype.names
            assert "truth_mask" in f["object_masks"].dtype.names


def run_onnx(train_dir, args=None):
    ckpt_path = [f for f in (train_dir / "ckpts").iterdir() if f.suffix == ".ckpt"][-1]
    if args is None:
        args = []
    args += [f"--ckpt_path={ckpt_path}"]
    args += ["--track_selection=dipsLoose202102"]

    to_onnx(args)
    get_onnx_metadata([str(train_dir / "network.onnx")])


def run_combined(
    tmp_path,
    config,
    do_eval=True,
    do_onnx=True,
    train_args=None,
    export_args=None,
    do_xbb=False,
    do_muP=False,
    inc_params=False,
):
    sys.argv = [sys.argv[0]]  # ignore pytest cli args when running salt cli

    # look for the config
    config_path = Path(__file__).parent.parent / "configs" / config
    if not config_path.is_file():
        config_path = Path(__file__).parent / "configs" / config

    # run training
    run_train(tmp_path, config_path, train_args, do_xbb, do_muP, inc_params)

    if do_eval:
        train_dir = [x for x in tmp_path.iterdir() if x.is_dir() and (x / "config.yaml").exists()]
        assert len(train_dir) == 1
        train_dir = train_dir[0]
        print(f"Using train_dir {train_dir}.")
        train_config_path = train_dir / "config.yaml"
        nd_path = [x for x in train_dir.iterdir() if x.suffix == ".yaml" and "norm" in str(x)]
        assert len(nd_path) == 1
        nd_path = nd_path[0]
        run_eval(tmp_path, train_config_path, nd_path, do_xbb)
    if do_onnx:
        run_onnx(train_dir, export_args)


@pytest.mark.filterwarnings(w)
def test_GN1(tmp_path) -> None:
    run_combined(tmp_path, "GN1.yaml", do_onnx=False)


@pytest.mark.filterwarnings(w)
def test_GN2(tmp_path) -> None:
    run_combined(tmp_path, CONFIG, do_onnx=False)


@pytest.mark.filterwarnings(w)
def test_GN3(tmp_path) -> None:
    run_combined(
        tmp_path,
        "GN3_dev/GN3_baseline.yaml",
        export_args=["--tasks", "jets_classification", "track_vertexing", "track_origin"],
    )


@pytest.mark.filterwarnings(w)
def test_GN2_muP(tmp_path) -> None:
    run_combined(tmp_path, "GN2_muP.yaml", do_muP=True, do_onnx=False)


@pytest.mark.filterwarnings(w)
def test_GN2emu(tmp_path) -> None:
    run_combined(tmp_path, "GN2emu.yaml", do_onnx=False)


@pytest.mark.filterwarnings(w)
def test_GN2XE(tmp_path) -> None:
    run_combined(tmp_path, "GN2XE.yaml", do_onnx=False, do_xbb=True)


@pytest.mark.filterwarnings(w)
def test_GN1_GATv2(tmp_path) -> None:
    args = [f"--config={Path(__file__).parent.parent / 'configs' / 'GATv2.yaml'}"]
    run_combined(tmp_path, "GN1.yaml", train_args=args, do_onnx=False)


@pytest.mark.filterwarnings(w)
def test_DIPS(tmp_path) -> None:
    run_combined(tmp_path, "dips.yaml", do_eval=True, do_onnx=True)


@pytest.mark.filterwarnings(w)
def test_DL1(tmp_path) -> None:
    run_combined(tmp_path, "DL1.yaml", do_eval=True, do_onnx=False)


@pytest.mark.filterwarnings(w)
def test_regression(tmp_path) -> None:
    run_combined(tmp_path, "regression.yaml", do_eval=True, do_onnx=True)


@pytest.mark.filterwarnings(w)
def test_nan_regression(tmp_path) -> None:
    run_combined(tmp_path, "nan_regression.yaml", do_eval=True, do_onnx=False)


@pytest.mark.filterwarnings(w)
def test_regression_gaussian(tmp_path) -> None:
    run_combined(tmp_path, "regression_gaussian.yaml", do_eval=True, do_onnx=True)


@pytest.mark.filterwarnings(w)
def test_flow(tmp_path) -> None:
    run_combined(tmp_path, "flow.yaml", do_eval=False, do_onnx=False)


@pytest.mark.filterwarnings(w)
def test_no_global_inputs(tmp_path) -> None:
    run_combined(tmp_path, "no_global_inputs.yaml", do_eval=False, do_onnx=False)


@pytest.mark.filterwarnings(w)
def test_train_dev(tmp_path) -> None:
    args = ["--trainer.fast_dev_run=2"]
    run_combined(tmp_path, CONFIG, do_eval=False, do_onnx=False, train_args=args)


@pytest.mark.filterwarnings(w)
def test_train_movefilestemp(tmp_path) -> None:
    tmp_path = Path(tmp_path)
    move_path = tmp_path / "dev" / "shm"
    args = [f"--data.move_files_temp={move_path}"]
    run_combined(tmp_path, CONFIG, do_eval=False, do_onnx=False, train_args=args)
    assert not Path(move_path).exists()


@pytest.mark.filterwarnings(w)
def test_train_distributed(tmp_path) -> None:
    args = ["--trainer.devices=2", "--data.num_workers=2", "--model.lrs_config.pct_start=0.2"]
    run_combined(tmp_path, CONFIG, do_eval=False, do_onnx=False, train_args=args)


@pytest.mark.filterwarnings(w)
def test_truncate_inputs(tmp_path) -> None:
    args = ["--data.num_inputs.tracks=10"]
    run_combined(tmp_path, CONFIG, do_eval=True, do_onnx=False, train_args=args)


@pytest.mark.filterwarnings(w)
def test_truncate_inputs_error(tmp_path) -> None:
    args = ["--data.num_inputs.this_should_error=10"]
    with pytest.raises(ValueError, match="must be a subset of"):
        run_combined(tmp_path, CONFIG, do_eval=False, do_onnx=False, train_args=args)


@pytest.mark.filterwarnings(w)
def test_tfv2(tmp_path) -> None:
    args = [f"--config={Path(__file__).parent.parent / 'configs' / 'encoder-v2.yaml'}"]
    run_combined(tmp_path, CONFIG, train_args=args)


@pytest.mark.filterwarnings(w)
def test_maskformer(tmp_path) -> None:
    run_combined(tmp_path, "MaskFormer.yaml", train_args=None, export_args=["-mf=vertexing"])


@pytest.mark.filterwarnings(w)
def test_param_concat(tmp_path) -> None:
    args = [f"--config={Path(__file__).parent.parent / 'tests' / 'configs' / 'param_concat.yaml'}"]
    run_combined(tmp_path, CONFIG, do_onnx=False, inc_params=True, train_args=args)


@pytest.mark.filterwarnings(w)
def test_param_featurewise(tmp_path) -> None:
    args = [
        f"--config={Path(__file__).parent.parent / 'tests' / 'configs' / 'param_featurewise.yaml'}"
    ]
    run_combined(tmp_path, CONFIG, do_onnx=False, inc_params=True, train_args=args)


@pytest.mark.filterwarnings(w)
def test_gls_weighting(tmp_path) -> None:
    # Ensure that this raises an assertion
    args = ["--model.loss_mode=lol"]
    with pytest.raises(AssertionError):
        run_combined(tmp_path, "dips.yaml", train_args=args)

    # Should fail, as we still have weights here
    args = ["--model.loss_mode=GLS"]
    with pytest.raises(AssertionError):
        run_combined(tmp_path, "GN2.yaml", train_args=args)

    # And this *should* work
    run_combined(tmp_path, "dips.yaml", train_args=args)



from collections import defaultdict
from pathlib import Path

import numpy as np
from ftag.hdf5 import H5Writer
from lightning import Callback, LightningModule, Trainer
from numpy.lib.recfunctions import unstructured_to_structured as u2s

from salt.models.task import (
    ClassificationTask,
    GaussianRegressionTask,
    RegressionTask,
    VertexingTask,
)
from salt.stypes import Vars
from salt.utils.array_utils import join_structured_arrays, maybe_pad
from salt.utils.mask_utils import indices_from_mask


class PredictionWriter(Callback):
    def __init__(
        self,
        write_tracks: bool = False,
        write_objects: bool = False,
        half_precision: bool = False,
        object_classes: list | None = None,
        extra_vars: Vars | None = None,
        global_object: str = "jets",
    ) -> None:
        """Write test outputs to h5 file.

        This callback will write the outputs of the model to an h5 evaluation file. The outputs
        are produced by calling the `run_inference` method of each task. The output file
        is written to the same directory as the checkpoint file, and has the same name
        as the checkpoint file, but with the suffix `__test_<sample><suffix>.h5`. The file will
        contain one dataset for each input type, with the same name as the input type in the test
        file.

        Parameters
        ----------
        write_tracks : bool
            If False, skip any tasks with `"tracks" in input_name`.
        write_objects : bool
            If False, skip any tasks with `input_name="objects"` and outputs of the
            MaskDecoder. Default is False
        half_precision : bool
            If true, write outputs at half precision
        object_classes : list
            List of flavour names with the index corresponding to the label values. This is used
            to construct the global object classification probability output names.
        extra_vars : Vars
            Extra variables to write to file for each input type. If not specified for a given input
            type, all variables in the test file will be written.
        global_object : str
            Name of the global object the model is trained on (usually the target for
            classification). By default "jets".
        """
        super().__init__()
        if extra_vars is None:
            extra_vars = defaultdict(list)
        self.extra_vars = extra_vars
        self.write_tracks = write_tracks
        self.write_objects = write_objects
        self.half_precision = half_precision
        self.precision = "f2" if self.half_precision else "f4"
        self.object_classes = object_classes
        self.global_object = global_object

    def setup(self, trainer: Trainer, module: LightningModule, stage: str) -> None:
        if stage != "test":
            return

        self.writer = None
        # some basic info
        self.trainer = trainer
        self.ds = trainer.datamodule.test_dataloader().dataset
        self.batch_size = trainer.datamodule.batch_size
        self.test_suff = trainer.datamodule.test_suff
        self.file = self.ds.file
        self.num = len(self.ds)
        self.norm_dict = self.ds.norm_dict

        # inputs names
        self.input_map = self.ds.input_map

        # check extra vars exist
        for input_type, vars_to_check in self.extra_vars.items():
            if input_type in self.input_map:
                dataset_name = self.input_map[input_type]
                available_vars = set(self.file[dataset_name].dtype.names)
                if missing_vars := set(vars_to_check) - available_vars:
                    raise ValueError(
                        "The following variables are missing for input type"
                        f"'{input_type}': {missing_vars}"
                    )
            else:
                raise ValueError(f"Input type '{input_type}' is not recognized in input_map.")

        # place to store intermediate outputs
        self.tasks = module.model.tasks
        self.outputs: dict = {input_name: {} for input_name in {t.input_name for t in self.tasks}}
        if self.extra_vars:
            self.outputs.update({input_name: {} for input_name in self.extra_vars})
        self.pad_masks: dict = {}

        # reformat output names for the global object classification task
        for task in self.tasks:
            if self.object_classes and task.name == f"{module.global_object}_classification":
                task.class_names = self.object_classes

        if self.write_objects:
            if not module.model.mask_decoder:
                raise ValueError("write_objects=True but no mask decoder found in model.")
            if not self.write_tracks:
                print("-" * 50)
                print("WARNING: If outputting mask objects, you probably also want tracks")

            # Add objects to outputs if there are no main tasks for this
            if "objects" not in self.outputs:
                self.outputs["objects"] = {}

            self.object_params = {
                "class_label": self.ds.mf_config.object.class_label,
                "label_map": [f"p{name}" for name in self.ds.mf_config.object.class_names],
            }

    @property
    def output_path(self) -> Path:
        out_dir = Path(self.trainer.ckpt_path).parent
        out_basename = str(Path(self.trainer.ckpt_path).stem)
        stem = str(Path(self.ds.filename).stem)
        sample = split[3] if len(split := stem.split("_")) == 4 else stem
        suffix = f"_{self.test_suff}" if self.test_suff is not None else ""
        return Path(out_dir / f"{out_basename}__test_{sample}{suffix}.h5")

    def _write_batch_outputs(self, batch_outputs, pad_masks, batch_idx):
        to_write = {}
        blow = batch_idx * self.batch_size
        bhigh = (batch_idx + 1) * self.batch_size
        for input_name, outputs in batch_outputs.items():
            this_outputs = []

            if input_name in self.input_map:
                name = self.input_map[input_name]

                # get input variables
                input_variables = None
                if name in self.extra_vars:
                    input_variables = self.extra_vars[name]
                if not input_variables:
                    input_variables = self.file[name].dtype.names
                if name in self.file:
                    inputs = self.file[name].fields(input_variables)[blow:bhigh]
                    this_outputs.append(inputs)
            else:
                name = input_name
                inputs = None

            for preds in outputs.values():
                if inputs is not None:
                    this_outputs.append(maybe_pad(preds, inputs))
                else:
                    this_outputs.append(preds)

            # add mask if present
            if name in pad_masks:
                pad_mask = pad_masks[name].cpu()
                pad_mask = u2s(np.expand_dims(pad_mask, -1), dtype=np.dtype([("mask", "?")]))
                this_outputs.append(maybe_pad(pad_mask, inputs))

            to_write[name] = join_structured_arrays(this_outputs)

        # If the writer hasn't been created yet, create it now that we have the dtypes and shapes
        if self.writer is None:
            dtypes = {k: v.dtype for k, v in to_write.items()}
            shapes = {k: (self.num,) + v.shape[1:] for k, v in to_write.items()}
            self.writer = H5Writer(
                dst=self.output_path,
                dtypes=dtypes,
                shapes=shapes,
                shuffle=False,
                jets_name=self.global_object,
                precision="half" if self.half_precision else "full",
            )
        self.writer.write(to_write)

    def on_test_batch_end(self, trainer, module, outputs, batch, batch_idx):  # noqa: ARG002
        preds = outputs
        _, pad_masks, labels = batch
        add_mask = False
        to_write = {input_name: {} for input_name in {t.input_name for t in self.tasks}}
        if self.extra_vars:
            to_write.update({input_name: {} for input_name in self.extra_vars})
        out_pads = {}
        for task in self.tasks:
            if not self.write_tracks and "tracks" in task.input_name:
                to_write.pop(task.input_name, None)
                continue
            if not self.write_objects and task.input_name == "objects":
                to_write.pop(task.input_name, None)
                continue

            this_preds = preds[task.input_name][task.name]
            this_pad_masks = pad_masks.get(task.input_name)

            # Get the outputs in the correct format
            if isinstance(task, ClassificationTask | VertexingTask):
                this_preds = task.get_h5(this_preds, this_pad_masks)
            if isinstance(task, RegressionTask | GaussianRegressionTask):
                this_preds = task.get_h5(this_preds, labels)

            # Add the outputs to the dictionary
            if task.name not in to_write[task.input_name]:
                to_write[task.input_name][task.name] = []
            if isinstance(task, GaussianRegressionTask):
                to_write[task.input_name][task.name] = this_preds[0]
                to_write[task.input_name][task.name + "_stddev"] = this_preds[1]
            else:
                to_write[task.input_name][task.name] = this_preds

            if this_pad_masks is not None and add_mask is False:
                out_pads[task.input_name] = this_pad_masks
                add_mask = True

        if self.write_objects:
            # Generate the object outputs of the form (B, N) where N is the number of objects
            objects = outputs["objects"]

            probs_dtype = np.dtype([
                (f"{module.name}_{n}", self.precision) for n in self.object_params["label_map"]
            ])
            to_write["objects"]["object_class_probs"] = u2s(
                objects["class_probs"].cpu().float().numpy(), dtype=probs_dtype
            )
            to_write["objects"]["object_class_targets"] = u2s(
                labels["objects"][self.object_params["class_label"]].cpu().unsqueeze(-1).numpy(),
                dtype=np.dtype([("class_label", "i8")]),
            )

            # Write the mask indices to the tracks
            mask_indices = indices_from_mask(objects["masks"].cpu().sigmoid() > 0.5)
            dtype = np.dtype([(f"{module.name}_MaskIndex", "i8")])
            mask_indices = mask_indices.int().cpu().numpy()
            mask_indices = np.where(~this_pad_masks, mask_indices, -1)
            to_write["tracks"]["mask_index"] = u2s(np.expand_dims(mask_indices, -1), dtype)

            # Write the truth mask and mask logits to their own dset
            to_write["object_masks"] = {}
            to_write["object_masks"]["tgt_masks"] = u2s(
                labels["objects"]["masks"].cpu().unsqueeze(-1).numpy(),
                dtype=np.dtype([("truth_mask", "i8")]),
            )
            to_write["object_masks"]["mask_logits"] = u2s(
                objects["masks"].cpu().float().unsqueeze(-1).numpy(),
                dtype=np.dtype([("mask_logits", self.precision)]),
            )

        self._write_batch_outputs(to_write, out_pads, batch_idx)

    def on_test_end(self, trainer, module):  # noqa: ARG002
        if self.writer is not None:
            self.writer.close()



import pytest
import torch

from salt.utils.mask_utils import indices_from_mask, mask_from_indices


@pytest.fixture
def indices_1d():
    return torch.tensor([0, 1, 1])


@pytest.fixture
def indices_2d():
    return torch.tensor([[0, 1, 1], [0, 0, 1]])


@pytest.fixture
def mask_2d():
    return torch.tensor([[True, False, False], [False, True, True]])


@pytest.fixture
def mask_3d():
    return torch.tensor([
        [[True, False, False], [False, True, True]],
        [[True, True, False], [False, False, True]],
    ])


def test_mask_from_indices_1d(indices_1d, mask_2d):
    sparse_from_1d = mask_from_indices(indices_1d)
    assert torch.all(sparse_from_1d == mask_2d)


def test_mask_from_indices_2d(indices_2d, mask_3d):
    sparse_from_2d = mask_from_indices(indices_2d)
    assert torch.all(sparse_from_2d == mask_3d)


def test_mask_from_negative_indices():
    indices = torch.tensor([-1, 0, 0, 1, 2])
    out = mask_from_indices(indices, num_masks=3)
    expected = torch.tensor([
        [False, True, True, False, False],
        [False, False, False, True, False],
        [False, False, False, False, True],
    ])
    assert torch.all(out == expected)

    indices = torch.tensor([[-1, -1, -1], [0, 0, 1]])
    out = mask_from_indices(indices, num_masks=3)
    expected = torch.tensor([
        [[False, False, False], [False, False, False], [False, False, False]],
        [[True, True, False], [False, False, True], [False, False, False]],
    ])
    assert torch.all(out == expected)


def test_indices_from_mask_2d(mask_2d, indices_1d):
    indices = indices_from_mask(mask_2d)
    assert torch.all(indices == indices_1d)


def test_indices_from_mask_3d(mask_3d, indices_2d):
    indices = indices_from_mask(mask_3d)
    assert torch.all(indices == indices_2d)


def test_indices_from_mask_empty():
    mask = torch.tensor([
        [False, False],
        [False, False],
        [False, False],
    ])
    indices = indices_from_mask(mask)
    assert torch.all(indices == torch.tensor([-2, -2]))


def test_indices_from_mask_with_non_mask_item():
    mask = torch.tensor([
        [True, True, True, False, False, False],
        [False, False, False, True, True, False],
    ])
    indices = indices_from_mask(mask)
    assert torch.all(indices == torch.tensor([0, 0, 0, 1, 1, -2]))



import lightning as L
from torch.utils.data import DataLoader

import salt.utils.file_utils as fu
from salt.data.datasets import SaltDataset
from salt.data.samplers import RandomBatchSampler


class SaltDataModule(L.LightningDataModule):
    def __init__(
        self,
        train_file: str,
        val_file: str,
        batch_size: int,
        num_workers: int,
        num_train: int,
        num_val: int,
        num_test: int,
        move_files_temp: str | None = None,
        class_dict: str | None = None,
        test_file: str | None = None,
        test_suff: str | None = None,
        pin_memory: bool = True,
        config_S3: dict | None = None,
        **kwargs,
    ):
        """Datamodule wrapping a [`salt.data.SaltDataset`][salt.data.SaltDataset] for training,
        validation and testing.

        This datamodule will load data from h5 files. The training, validation and test files
        are specified by the `train_file`, `val_file` and `test_file` arguments.

        The arguments of this class can be set from the YAML config file or from the command line
        using the `data` key. For example, to set the `batch_size` from the command line, use
        `--data.batch_size=1000`.

        Parameters
        ----------
        train_file : str
            Training file path
        val_file : str
            Validation file path
        batch_size : int
            Number of samples to process in each training step
        num_workers : int
            Number of CPU worker processes to load batches from disk
        num_train : int
            Total number of training samples
        num_val : int
            Total number of validation samples
        num_test : int
            Total number of testing samples
        move_files_temp : str
            Directory to move training files to, default is None,
            which will result in no copying of files
        class_dict : str
            Path to umami preprocessing scale dict file
        test_file : str
            Test file path, default is None
        test_suff : str
            Test file suffix, default is None
        pin_memory: bool
            Pin memory for faster GPU transfer, default is True
        config_S3: dict, optional
            Some parameters for the S3 access
        **kwargs
            Keyword arguments for [`salt.data.SaltDataset`][salt.data.SaltDataset]
        """
        super().__init__()
        self.train_file = train_file
        self.val_file = val_file
        self.test_file = test_file
        self.test_suff = test_suff
        self.batch_size = batch_size
        self.num_workers = num_workers
        self.num_train = num_train
        self.num_val = num_val
        self.num_test = num_test
        self.class_dict = class_dict
        self.move_files_temp = move_files_temp
        self.pin_memory = pin_memory
        self.config_S3 = config_S3
        self.kwargs = kwargs

    def prepare_data(self):
        if self.move_files_temp and not self.trainer.fast_dev_run:
            print("-" * 100)
            print(f"Moving train files to {self.move_files_temp} ")
            print("-" * 100)
            fu.move_files_temp(self.move_files_temp, self.train_file, self.val_file)

    def setup(self, stage: str):
        if self.trainer is not None and self.trainer.is_global_zero:
            print("-" * 100)

        if stage == "fit" and self.move_files_temp and not self.trainer.fast_dev_run:
            # Set the training/validation file to the temp path
            self.train_file = fu.get_temp_path(self.move_files_temp, self.train_file)
            self.val_file = fu.get_temp_path(self.move_files_temp, self.val_file)

        # create training and validation datasets
        if stage == "fit":
            self.train_dset = SaltDataset(
                filename=self.train_file,
                num=self.num_train,
                stage=stage,
                **self.kwargs,
            )
            self.val_dset = SaltDataset(
                filename=self.val_file,
                num=self.num_val,
                stage=stage,
                **self.kwargs,
            )

        # Only print train/val dataset details when actually training
        if stage == "fit" and self.trainer.is_global_zero:
            print(f"Created training dataset with {len(self.train_dset):,} entries")
            print(f"Created validation dataset with {len(self.val_dset):,} entries")

        if stage == "test":
            assert self.test_file is not None, "No test file specified, see --data.test_file"
            self.test_dset = SaltDataset(
                filename=self.test_file,
                num=self.num_test,
                stage=stage,
                **self.kwargs,
            )
            print(f"Created test dataset with {len(self.test_dset):,} entries")

        if self.trainer.is_global_zero:
            print("-" * 100, "\n")

    def get_dataloader(self, stage: str, dataset: SaltDataset, shuffle: bool):
        drop_last = stage == "fit"
        return DataLoader(
            dataset=dataset,
            batch_size=None,
            collate_fn=None,
            sampler=RandomBatchSampler(dataset, self.batch_size, shuffle, drop_last=drop_last),
            num_workers=self.num_workers,
            shuffle=False,
            pin_memory=self.pin_memory,
        )

    def train_dataloader(self):
        return self.get_dataloader(dataset=self.train_dset, stage="fit", shuffle=True)

    def val_dataloader(self):
        return self.get_dataloader(dataset=self.val_dset, stage="test", shuffle=False)

    def test_dataloader(self):
        return self.get_dataloader(dataset=self.test_dset, stage="test", shuffle=False)

    def teardown(self, stage: str | None = None):
        if (
            stage == "fit"
            and self.move_files_temp
            and not self.trainer.fast_dev_run
            and self.trainer.is_global_zero
        ):
            print("-" * 100)
            print(f"Removing training files: \n\t{self.train_file}\n\t{self.val_file}")
            fu.remove_files_temp(self.train_file, self.val_file)
            print("-" * 100)



import torch
import torch.nn.functional as nnf
import torchmetrics as tm
from lightning import Callback, LightningModule, Trainer
from torch.nn import Module, ModuleDict

from salt.utils.mask_utils import mask_from_logits, reco_metrics


class MaskformerMetrics(Callback):
    def __init__(self, only_val=True, mask_criteria: dict[str, tuple] | None = None):
        """Callback to log metrics for Maskformer model.

        The following metrics are logged.
            - class_exact_match
            - not null class efficiency and purity
            - per class efficiency and purity
            - mean absolute error for regression tasks
            - mask match efficiency and fake rate, for different criteria

        Parameters
        ----------
        only_val: bool
            Whether to log metrics only on validation step. Default is true. It is not recommended
            to log metrics on training step, as it will slow down the training process.
        mask_criteria: dict[str, tuple] | None
            Dictionary with the criteria to evaluate the mask match. The keys are the name of the
            criteria, and the values are tuples (R, P) where R is the mimimum recall and P is the
            minimum purity for the criteria to be met. Default is None, which will define the
            'perfect' critera as (1, 1) and the 'loose' criteria as (0.5, 0.5).

        """
        self.mask_criteria = mask_criteria
        if not self.mask_criteria:
            self.mask_criteria = {"perfect": (1, 1), "loose": (0.5, 0.5)}
        self.only_val = only_val

    def setup(self, trainer: Trainer, module: LightningModule, stage: str) -> None:
        if trainer.fast_dev_run or stage != "fit":
            return
        model = module.model
        if not model.mask_decoder:
            raise ValueError("Model requires mask_decoder to use MaskformerMetrics callback")
        kwargs = {"sync_dist": len(trainer.device_ids) > 1}

        def log(metrics, stage):
            for t, loss_value in metrics.items():
                module.log(f"{stage}/{t}", loss_value, **kwargs)

        self.log = log
        # make Not hard-coded
        self.classes = trainer.datamodule.train_dataloader().dataset.mf_config.object.class_names
        self.null_index = len(self.classes) - 1
        nclasses = len(self.classes)
        mask_size = 40

        module.maskformer_metrics = Module()
        self.metrics = module.maskformer_metrics
        # class label metrics
        self.metrics.class_exact_match = tm.classification.MulticlassExactMatch(
            num_classes=nclasses
        )
        self.metrics.class_accuracy_micro = tm.classification.MulticlassAccuracy(
            num_classes=nclasses, average="micro"
        )
        self.metrics.class_accuracy_macro = tm.classification.MulticlassAccuracy(
            num_classes=nclasses, average="macro"
        )

        # per-class metrics
        self.metrics.per_class_eff = ModuleDict({
            k: tm.classification.BinaryRecall() for k in self.classes if k != "null"
        })
        self.metrics.per_class_pur = ModuleDict({
            k: tm.classification.BinaryPrecision() for k in self.classes if k != "null"
        })
        # not-null metrics
        self.metrics.notnull_eff = tm.classification.BinaryRecall()
        self.metrics.notnull_pur = tm.classification.BinaryPrecision()

        # mask metrics
        self.metrics.mask_exact_match = tm.classification.MultilabelExactMatch(num_labels=mask_size)

    def _get_metrics(self, module, outputs):
        metrics = {}

        preds, labels, pad_mask = (outputs["preds"], outputs["labels"], outputs["pad_masks"])
        qpreds, qlabels = preds["objects"], labels["objects"]

        # class metrics
        obj_class_pred = qpreds["class_logits"].argmax(-1).detach()
        obj_class_tgt = qlabels["object_class"].detach()

        self.metrics.class_exact_match(obj_class_pred, obj_class_tgt)
        metrics["class_exact_match"] = self.metrics.class_exact_match
        self.metrics.class_accuracy_micro(obj_class_pred.view(-1), obj_class_tgt.view(-1))
        metrics["class_accuracy_micro"] = self.metrics.class_accuracy_micro
        self.metrics.class_accuracy_macro(obj_class_pred.view(-1), obj_class_tgt.view(-1))
        metrics["class_accuracy_macro"] = self.metrics.class_accuracy_macro

        # per-class metrics
        for i, k in enumerate(self.classes):
            if k == "null":
                obj_present_tgt = obj_class_tgt != self.null_index
                obj_present_pred = obj_class_pred != self.null_index
                self.metrics.notnull_eff(obj_present_pred, obj_present_tgt)
                self.metrics.notnull_pur(obj_present_pred, obj_present_tgt)
                metrics["notnull_eff"] = self.metrics.notnull_eff
                metrics["notnull_pur"] = self.metrics.notnull_pur
            else:
                is_class_tgt = obj_class_tgt.view(-1) == i
                is_class_pred = obj_class_pred.view(-1) == i
                self.metrics.per_class_eff[k](is_class_pred, is_class_tgt)
                self.metrics.per_class_pur[k](is_class_pred, is_class_tgt)
                metrics[f"{k}_eff"] = self.metrics.per_class_eff[k]
                metrics[f"{k}_pur"] = self.metrics.per_class_pur[k]

        pred_masks = qpreds["masks"].detach()
        tgt_masks = qlabels["masks"].detach()
        pred_masks = mask_from_logits(pred_masks, "sigmoid", pad_mask["tracks"], obj_class_pred)

        for key, (recall, purity) in self.mask_criteria.items():
            eff, fr = reco_metrics(
                pred_masks, tgt_masks, min_recall=recall, min_purity=purity, reduce=True
            )
            metrics[f"query_{key}_match_eff"] = eff
            metrics[f"query_{key}_match_fake"] = fr

        if "regression" in qpreds:
            task = [
                t
                for t in module.model.tasks
                if t.input_name == "objects" and t.name == "regression"
            ]
            if len(task) != 1:
                raise ValueError(
                    "Can only run regression maskformer metrics for a single query regression task"
                )
            task = task[0]

            valid_idx = ~torch.isnan(qlabels["regression"]).all(-1)
            for i, t in enumerate(task.targets):
                unscaled_preds = task.scaler.inverse(t, qpreds["regression"][valid_idx][:, i])
                targets = qlabels[t][valid_idx]
                metrics[f"query_{t}_mae"] = nnf.l1_loss(unscaled_preds, targets)

        return metrics

    def _get_log_metrics(self, module, outputs, stage):
        metrics = self._get_metrics(module, outputs["outputs"])
        self.log(metrics, stage)

    def on_train_batch_end(self, trainer, module, outputs, batch, batch_idx):  # noqa: ARG002
        if not self.only_val and not trainer.fast_dev_run:
            self._get_log_metrics(module, outputs, "train")

    def on_validation_batch_end(self, trainer, module, outputs, batch, batch_idx):  # noqa: ARG002
        if not trainer.fast_dev_run:
            self._get_log_metrics(module, outputs, "val")



from pathlib import Path

from lightning import LightningModule, Trainer
from lightning.pytorch.callbacks import ModelCheckpoint


class Checkpoint(ModelCheckpoint):
    def __init__(self, monitor_loss: str, fname_string: str = "val_loss") -> None:
        """Create checkpoint files during training."""
        filename = "epoch={epoch:03d}-" + fname_string + "={" + monitor_loss + ":.5f}"
        super().__init__(save_top_k=-1, filename=filename, auto_insert_metric_name=False)

    def setup(self, trainer: Trainer, pl_module: LightningModule, stage: str) -> None:
        if stage == "fit":
            if trainer.fast_dev_run:
                return

            if "s3:/" in trainer.log_dir[:4]:
                from s3path import S3Path

                log_dir = S3Path(trainer.log_dir.replace("s3://", "").replace("s3:/", ""))
                self.dirpath = "s3://" + str(log_dir / "ckpts")
            elif "s3:/" in trainer.log_dir:
                raise ValueError(
                    f"trainer.log_dir should start with 's3:/', instead of {trainer.log_dir}"
                )
            else:
                log_dir = Path(trainer.log_dir)
                self.dirpath = str(log_dir / "ckpts")

            # this could be used to add the timestamp to the filename
            # self.timestamp = log_dir.name

        super().setup(trainer=trainer, pl_module=pl_module, stage=stage)



import torch
from torch import Tensor, cat, nn

from salt.models import FeaturewiseTransformation, InitNet, Pooling
from salt.stypes import BoolTensors, NestedTensors, Tensors
from salt.utils.tensor_utils import flatten_tensor_dict, maybe_flatten_tensors


class SaltModel(nn.Module):
    def __init__(
        self,
        init_nets: list[dict],
        tasks: nn.ModuleList,
        encoder: nn.Module = None,
        mask_decoder: nn.Module = None,
        pool_net: Pooling = None,
        merge_dict: dict[str, list[str]] | None = None,
        featurewise_nets: list[dict] | None = None,
    ):
        """A generic multi-modal, multi-task neural network.

        This model can be used to implement a wide range of models, including
        [DL1](https://ftag.docs.cern.ch/algorithms/taggers/dips/),
        [DIPS](https://ftag.docs.cern.ch/algorithms/taggers/dl1/),
        [GN2](https://ftag.docs.cern.ch/algorithms/taggers/GN2/)
        and more.

        Parameters
        ----------
        init_nets : list[dict]
            Keyword arguments for one or more initialisation networks.
            See [`salt.models.InitNet`][salt.models.InitNet].
            Each initialisation network produces an initial input embedding for
            a single input type.
        tasks : nn.ModuleList
            Task heads, see [`salt.models.TaskBase`][salt.models.TaskBase].
            These can be used to implement object tagging, vertexing, regression,
            classification, etc.
        encoder : nn.Module
            Main input encoder, which takes the output of the initialisation
            networks and produces a single embedding for each constituent.
            If not provided this model is essentially a DeepSets model.
        mask_decoder : nn.Module
            Mask decoder, which takes the output of the encoder and produces a
            series of learned embeddings to represent object masks
        pool_net : nn.Module
            Pooling network which computes a global representation of the object
            by aggregating over the constituents. If not provided, assume that
            the only inputs are global features (i.e. no constituents).
        merge_dict : dict[str, list[str]] | None
            A dictionary that lets the salt concatenate all the input
            representations of the inputs in list[str] and act on them
            in following layers (e.g. transformer or tasks) as if they
            are coming from one input type
        featurewise_nets : list[dict]
            Keyword arguments for featurewise transformation networks that perform
            featurewise scaling and biasing.
        """
        super().__init__()

        # init featurewise networks
        if featurewise_nets:
            self.init_featurewise(featurewise_nets, init_nets, encoder)

        self.init_nets = nn.ModuleList([InitNet(**init_net) for init_net in init_nets])
        self.tasks = tasks
        self.encoder = encoder
        self.mask_decoder = mask_decoder

        self.pool_net = pool_net
        self.merge_dict = merge_dict

        # checks for the global object only setup
        if self.pool_net is None:
            assert self.encoder is None, "pool_net must be set if encoder is set"
            assert len(self.init_nets) == 1, "pool_net must be set if more than one init_net is set"
            assert self.init_nets[0].input_name == self.init_nets[0].global_object

    def forward(
        self,
        inputs: Tensors,
        pad_masks: BoolTensors | None = None,
        labels: NestedTensors | None = None,
    ) -> tuple[NestedTensors, Tensors]:
        """Forward pass through the `SaltModel`.

        Don't call this method directy, instead use `__call__`.

        Parameters
        ----------
        inputs : Tensors
            Dict of input tensors for each modality. Each tensor is of shape
            `(batch_size, num_inputs, input_size)`.
        pad_masks : BoolTensors
            Dict of input padding mask tensors for each modality. Each tensor is of
            shape `(batch_size, num_inputs)`.
        labels : Tensors, optional
            Nested dict of label tensors. The outer dict is keyed by input modality,
            the inner dict is keyed by label variable name. Each tensor is of shape
            `(batch_size, num_inputs)`. If not specified, assume we are running model
            inference (i.e. no loss computation).

        Returns
        -------
        preds : NestedTensors
            Dict of model predictions for each task, separated by input modality.
            Tensors have varying shapes depending on the task.
        loss : Tensors
            Dict of losses for each task, aggregated over the batch.
        """
        # initial input projections
        xs = {}

        for init_net in self.init_nets:
            xs[init_net.input_name] = init_net(inputs)

        # handle edge features if present
        edge_x = xs.pop("EDGE", None)
        kwargs = {} if edge_x is None else {"edge_x": edge_x}

        if isinstance(self.merge_dict, dict):
            for merge_name, merge_types in self.merge_dict.items():
                xs[merge_name] = cat([xs.pop(mt) for mt in merge_types], dim=1)
            if pad_masks is not None:
                for merge_name, merge_types in self.merge_dict.items():
                    pad_masks[merge_name] = cat([pad_masks.pop(mt) for mt in merge_types], dim=1)
            for merge_name, merge_types in self.merge_dict.items():
                if labels is not None:
                    labels[merge_name] = {}
                    for var in labels[merge_types[0]]:
                        labels[merge_name].update({
                            var: cat([labels[mt][var] for mt in merge_types], dim=1)
                        })

        # Generate embedding from encoder, or by concatenating the init net outputs
        # We should change this such that all encoders return (x, mask)
        if self.encoder:
            embed_xs = self.encoder(xs, pad_mask=pad_masks, inputs=inputs, **kwargs)
            if isinstance(embed_xs, tuple):
                embed_xs, pad_masks = embed_xs
            preds = {"embed_xs": embed_xs}
        else:
            preds = {"embed_xs": flatten_tensor_dict(xs)}

        preds, labels, loss = (
            self.mask_decoder(preds, self.tasks, pad_masks, labels)
            if self.mask_decoder
            else (preds, labels, {})
        )

        # apply featurewise transformation to global track embeddings if configured
        if hasattr(self, "featurewise_global") and self.featurewise_global:
            preds["embed_xs"] = self.featurewise_global(inputs, preds["embed_xs"])

        # pooling
        if self.pool_net:
            global_rep = self.pool_net(preds, pad_mask=pad_masks)
        else:
            global_rep = preds["embed_xs"]

        # add global features to global representation
        if (global_feats := inputs.get("GLOBAL")) is not None:
            global_rep = torch.cat([global_rep, global_feats], dim=-1)
        preds["global_rep"] = global_rep

        # run tasks
        task_preds, task_loss = self.run_tasks(preds, pad_masks, labels)
        preds.update(task_preds)
        loss.update(task_loss)

        return preds, loss

    def run_tasks(
        self,
        preds: dict[str, Tensor],
        masks: BoolTensors | None,
        labels: NestedTensors | None = None,
    ):
        loss = {}

        preds["embed_xs"] = maybe_flatten_tensors(preds["embed_xs"])

        for task in self.tasks:
            if task.input_name == task.global_object:
                task_preds, task_loss = task(preds["global_rep"], labels, None, context=None)
            elif task.input_name == "objects":
                task_preds, task_loss = task(preds["objects"]["embed"], labels, masks, context=None)
            else:
                task_preds, task_loss = task(
                    preds["embed_xs"], labels, masks, context=preds["global_rep"]
                )
            if task.input_name not in preds:
                preds[task.input_name] = {}
            preds[task.input_name][task.name] = task_preds
            loss[task.name] = task_loss

        return preds, loss

    def init_featurewise(
        self, featurewise_nets: list[dict], init_nets: list[dict], encoder: nn.Module
    ):
        for featurewise_net in featurewise_nets:
            if featurewise_net.get("layer") == "input":
                for init_net in init_nets:
                    init_net["featurewise"] = FeaturewiseTransformation(**featurewise_net)
            elif featurewise_net.get("layer") == "encoder":
                if encoder:
                    for _layer in range(encoder.num_layers):
                        encoder.featurewise.append(FeaturewiseTransformation(**featurewise_net))
                else:
                    raise ValueError(
                        "Requested featurewise transforms for encoder, no encoder configured"
                    )
            elif featurewise_net.get("layer") == "global":
                self.featurewise_global = FeaturewiseTransformation(**featurewise_net)
            else:
                raise ValueError(
                    "Select either 'input', 'encoder' or 'global' layers for featurewise nets."
                )



"""Efficient Transformer implementation.

Updated transformer implementation based on
https://github.com/mistralai/mistral-src

Features:
- native SDP kernels (including flash)
- gated linear units https://arxiv.org/abs/2002.05202
- RMSNorm https://arxiv.org/abs/1910.07467
"""

import math
import warnings
from functools import partial

import torch
import torch.nn.functional as F
from torch import BoolTensor, Tensor, nn

import salt.models.layernorm as layernorms
from salt.stypes import Tensors
from salt.utils.tensor_utils import redo_padding, undo_padding

ATTN_TYPES = ["torch-math", "torch-flash", "torch-meff", "flash-varlen"]


def merge_masks(
    kv_mask: BoolTensor | None,
    attn_mask: BoolTensor | None,
    q_shape: Tensor,
) -> BoolTensor | None:
    """Create a full attention mask which incorporates the padding information.

    Using pytorch transformer convention for padding
        False: Real node
        True:  Zero padded

    Using pytorch transformer convention for attention mask
        False:  Not allowed in attention mechanism
        True:   Allowed in attention mechanism

    Designing attention mask such that padded tokens can't send information.
    But they can receive them.
    This prevents Nans in the attention scores caused by the softmax

    Parameters
    ----------
    kv_mask : BoolTensor | None
        Mask for the keys and values, of shape (batch, kv_len).
    attn_mask : BoolTensor | None
        Full attention mask, of shape (batch, q_len, kv_len).
    q_shape : Size
        Shape of the queries tensor, (batch, q_len, dim).
    """
    # Create the full mask which combines the attention and padding masks
    mask = None

    # if the kv_mask mask exists, ensure that padded tokens never send information
    if kv_mask is not None:
        mask = kv_mask.unsqueeze(-2).expand(-1, q_shape[-2], -1)
        mask = ~mask  # convert the mask such that True is a valid token

    # include the attention mask
    if attn_mask is not None:
        mask = attn_mask if mask is None else attn_mask & mask

    # Unsqueeze the mask to give it a dimension for num_head broadcasting
    if mask is not None:
        mask = mask.unsqueeze(1)

    return mask


def repeat_kv(keys: Tensor, values: Tensor, repeats: int, dim: int):
    keys = torch.repeat_interleave(keys, repeats=repeats, dim=dim)
    values = torch.repeat_interleave(values, repeats=repeats, dim=dim)
    return keys, values


def change_attn_backends(module: nn.Module, backend: str) -> None:
    """Recursively change the attention backend of a module and all its children.

    Used primarily for switching back to torch-math for ONNX exports.
    """
    if isinstance(module, TransformerV2):
        module.set_backend(backend)
        return
    if isinstance(module, Attention):
        module.set_backend(backend)
        return
    for child in module.children():
        change_attn_backends(child, backend)


def projection_packed(
    q: Tensor,
    kv: Tensor | None,
    weight: Tensor,
    bias: Tensor | None = None,
) -> tuple:
    """Efficient input projection for MHA when using a single linear layer.

    Essentially the same as torch.nn.functional._in_projection_packed
    But here we use chunk which is 40x faster than unflatten
    Not sure why they don't use chunk in the original implementation...

    Parameters
    ----------
    q : Tensor
        The queries tensor of shape (batch, q_len, dim).
    kv : Tensor | None
        The keys and values tensor of shape (batch, kv_len, dim).
    weight : Tensor
        The packed weight tensor of the input lienar projection with shape (3 * dim, dim).
    bias : Tensor | None
        The optional packed bias tensor of the input linear projection with shape (3 * dim).

    Returns
    -------
    q_proj, k_proj, v_proj : tuple
        The projected queries, keys, and values tensors.
    """
    # If the q tensor is the only input, then we assume we are doing self-attention.
    # This is made (slightly) faster by using a single linear layer, then chunking rather than
    # three seperate linear layers processed one at a time.
    if kv is None:
        return F.linear(q, weight, bias).chunk(3, dim=-1)

    # If the kv tensor is present, then we are doing cross-attention.
    # This means we must project the q and kv tensors seperately.
    # The kv linear layer can remain packed, allowing us to project together then chunk,
    # using the same trick as above. We must however first seperate weights (and biases if present)
    # of the linear layers for the q and kv parts. We use torch.split which returns a veiw of the
    # original tensor so this step doesnt required any extra memory or much time.
    dim = q.size(-1)
    w_q, w_kv = weight.split([dim, dim * 2])
    b_q, b_kv = bias.split([dim, dim * 2]) if bias is not None else (None, None)

    # Now we can do the seperate projections
    q_proj = F.linear(q, w_q, b_q)
    k_proj, v_proj = F.linear(kv, w_kv, b_kv).chunk(2, dim=-1)
    return q_proj, k_proj, v_proj


def torch_attn(
    q: Tensor, k: Tensor, v: Tensor, mask: BoolTensor, dropout: float, backend: str
) -> Tensor:
    """Torch dot product attention with a switchable backend."""
    with torch.backends.cuda.sdp_kernel(
        enable_math=True,  # always enabled as a fallback
        enable_mem_efficient=(backend == "torch-meff"),
        enable_flash=(backend == "torch-flash"),
    ):
        return F.scaled_dot_product_attention(q, k, v, attn_mask=mask, dropout_p=dropout)


class Attention(nn.Module):
    def __init__(
        self,
        embed_dim: int,
        num_heads: int = 1,
        attn_type: str = "torch-meff",
        dropout: float = 0.0,
        bias: bool = True,
        diff_attention: bool = False,
        depth: int = 1,
        do_qk_norm: bool = False,
        do_v_norm: bool = False,
    ) -> None:
        """Multihead attention module.

        Parameters
        ----------
        embed_dim : int
            Dimension of the input.
        num_heads : int
            Number of attention heads.
        attn_type : str, optional
            Name of backend kernel to use.
        dropout : float, optional
            Dropout rate.
        bias : bool, optional
            Whether to include bias terms.
        diff_attention : bool, optional
            Use differential attention or not
        depth : int, optional
            Number of current attention layer
        do_qk_norm : bool, optional
            Whether to apply norm to q and k
        do_v_norm : bool, optional
            Whether to apply norm to v
        """
        super().__init__()
        assert embed_dim % num_heads == 0, "Dim not div by the number of heads!"
        assert attn_type in ATTN_TYPES, "Invalid attention type!"

        # Attributes
        self.embed_dim = embed_dim
        self.num_heads = num_heads
        self.head_dim = embed_dim // num_heads
        self.dropout = dropout
        self.bias = bias
        self.attn_type = attn_type
        self.diff_attention = diff_attention
        self.depth = depth
        self.do_qk_norm = do_qk_norm
        self.do_v_norm = do_v_norm

        if self.diff_attention:
            self.head_dim = self.head_dim // 2
            self.lambda_init = 0.8 - 0.6 * math.exp(-0.3 * depth)
            self.lambda_q1 = nn.Parameter(
                torch.zeros(self.head_dim, dtype=torch.float32).normal_(mean=0, std=0.1)
            )
            self.lambda_k1 = nn.Parameter(
                torch.zeros(self.head_dim, dtype=torch.float32).normal_(mean=0, std=0.1)
            )
            self.lambda_q2 = nn.Parameter(
                torch.zeros(self.head_dim, dtype=torch.float32).normal_(mean=0, std=0.1)
            )
            self.lambda_k2 = nn.Parameter(
                torch.zeros(self.head_dim, dtype=torch.float32).normal_(mean=0, std=0.1)
            )

            self.subln = layernorms.RMSNorm(2 * self.head_dim)

        if self.do_qk_norm:
            self.q_norm = layernorms.RMSNorm(self.head_dim)
            self.k_norm = layernorms.RMSNorm(self.head_dim)
        if self.do_v_norm:
            self.v_norm = layernorms.RMSNorm(self.head_dim)

        # Better parallelism for self-attention when using parameters directly
        self.in_proj_weight = nn.Parameter(torch.empty(3 * embed_dim, embed_dim))
        self.in_proj_bias = nn.Parameter(torch.empty(3 * embed_dim)) if bias else None
        self.out_proj = nn.Linear(embed_dim, embed_dim, bias=bias)
        self.reset_parameters()
        self.set_backend(attn_type)

    def set_backend(self, attn_type: str):
        # Check the attention backend
        self.attn_type = attn_type
        if self.attn_type == "flash-varlen":
            why_not_flash = ""
            try:
                from flash_attn import flash_attn_varlen_qkvpacked_func

                self._flash_attn = flash_attn_varlen_qkvpacked_func
            except ImportError:
                why_not_flash = (
                    "Requires the flash_attn package, CUDA 12+, and A100+, and must be installed "
                    "separately or using the [flash] extra. See requirements-flash.txt."
                )
            if not torch.cuda.is_available():
                why_not_flash = "No GPU available."
            if why_not_flash:
                warnings.warn(
                    f"Cannot use flash-varlen backend. {why_not_flash} Reverting to torch-math.",
                    stacklevel=2,
                )
                self.attn_type = "torch-math"
        return self.attn_type

    def reset_parameters(self):
        """Initialize the parameters."""
        nn.init.xavier_uniform_(self.in_proj_weight)
        if self.bias:
            nn.init.constant_(self.in_proj_bias, 0.0)
        self.out_proj.reset_parameters()

    def _weight_by_lambda(self, attn1: Tensor, attn2: Tensor) -> Tensor:
        lambda_1 = torch.exp(torch.sum(self.lambda_q1 * self.lambda_k1, dim=-1).float()).type_as(
            attn1
        )
        lambda_2 = torch.exp(torch.sum(self.lambda_q2 * self.lambda_k2, dim=-1).float()).type_as(
            attn1
        )
        lambda_full = lambda_1 - lambda_2 + self.lambda_init
        attn = attn1 - lambda_full * attn2

        return self.subln(attn) * (1 - self.lambda_init)

    def _flash_forward(self, x: Tensor, culens: Tensor, maxlen: int) -> Tensor:
        """FlashAttention backend."""
        # Perform the packed input projection
        qkv = F.linear(x, self.in_proj_weight, self.in_proj_bias)
        qkv = qkv.view(-1, 3, self.num_heads, self.head_dim)

        if self.do_qk_norm:
            dtype = qkv.dtype
            if self.do_v_norm:
                q, k, v = qkv.unbind(1)
                q = self.q_norm(q)
                k = self.k_norm(k)
                v = self.v_norm(v)
                qkv = torch.stack([q, k, v], dim=1).to(dtype)
            else:
                q, k, v = qkv.unbind(1)
                q = self.q_norm(q)
                k = self.k_norm(k)
                qkv = torch.stack([q, k, v], dim=1).to(dtype)

        # Run the flash-varlen backend
        dropout = self.dropout if self.training else 0.0
        a_out = self._flash_attn(qkv, culens, maxlen, dropout)
        a_out = a_out.reshape(-1, self.embed_dim)

        # Mix with final linear layer
        return self.out_proj(a_out)

    def _flash_diff_forward(self, x: Tensor, culens: Tensor, maxlen: int) -> Tensor:
        """FlashAttention backend."""
        # Perform the packed input projection
        qkv = F.linear(x, self.in_proj_weight, self.in_proj_bias)
        qkv = qkv.view(-1, 3, self.num_heads, 2, self.head_dim)

        q1, q2 = qkv[:, 0, :, 0], qkv[:, 0, :, 1]
        k1, k2 = qkv[:, 1, :, 0], qkv[:, 1, :, 1]
        v1, v2 = qkv[:, 2, :, 0], qkv[:, 2, :, 1]

        qkv11 = torch.stack([q1, k1, v1], dim=1)
        qkv12 = torch.stack([q1, k1, v2], dim=1)
        qkv21 = torch.stack([q2, k2, v1], dim=1)
        qkv22 = torch.stack([q2, k2, v2], dim=1)

        # Run the flash-varlen backend
        dropout = self.dropout if self.training else 0.0
        a_out11 = self._flash_attn(qkv11, culens, maxlen, dropout)
        a_out12 = self._flash_attn(qkv12, culens, maxlen, dropout)
        a_out1 = torch.cat([a_out11, a_out12], dim=-1)

        a_out21 = self._flash_attn(qkv21, culens, maxlen, dropout)
        a_out22 = self._flash_attn(qkv22, culens, maxlen, dropout)
        a_out2 = torch.cat([a_out21, a_out22], dim=-1)

        a_out = self._weight_by_lambda(a_out1, a_out2)
        a_out = a_out.reshape(-1, self.embed_dim)

        # Mix with final linear layer
        return self.out_proj(a_out)

    def _torch_forward(
        self, x: Tensor, kv: Tensor, mask: BoolTensor, kv_mask: BoolTensor, attn_mask: BoolTensor
    ) -> Tensor:
        """Attention using pytorch."""
        # Otherwise perform standard attention
        B, S, D = x.shape

        # input projections -> B, S, D
        q, k, v = projection_packed(x, kv, self.in_proj_weight, self.in_proj_bias)

        # transform tensors to (B, Nh, S, Hd)
        shape = (B, -1, self.num_heads, self.head_dim)  # Dont use S for cross attn
        q, k, v = (t.view(shape).transpose(1, 2).contiguous() for t in (q, k, v))

        if self.do_qk_norm:
            q = self.q_norm(q)
            k = self.k_norm(k)
        if self.do_v_norm:
            v = self.v_norm(v)

        # run attention
        s_mask = mask if kv is None else kv_mask  # Who is sending, x or kv
        mask = merge_masks(s_mask, attn_mask, q.shape)
        dropout = self.dropout if self.training else 0.0
        a_out = torch_attn(q, k, v, mask, dropout, self.attn_type)

        # recombine heads
        a_out = a_out.transpose(1, 2).contiguous().view(B, S, D)

        # mix with final linear layer
        return self.out_proj(a_out)

    def _torch_diff_forward(
        self, x: Tensor, kv: Tensor, mask: BoolTensor, kv_mask: BoolTensor, attn_mask: BoolTensor
    ) -> Tensor:
        """Attention using pytorch."""
        # Otherwise perform standard attention
        B, S, D = x.shape

        # input projections -> B, S, D
        q, k, v = projection_packed(x, kv, self.in_proj_weight, self.in_proj_bias)

        # transform tensors to (B, Nh, S, Hd)
        shape = (B, -1, self.num_heads, 2, self.head_dim)  # Dont use S for cross attn
        q, k, v = (t.view(shape).transpose(1, 2).contiguous() for t in (q, k, v))

        # Split q, k, v into two parts
        q1, q2 = q[:, :, :, 0], q[:, :, :, 1]
        k1, k2 = k[:, :, :, 0], k[:, :, :, 1]
        v1, v2 = v[:, :, :, 0], v[:, :, :, 1]

        # run attention
        s_mask = mask if kv is None else kv_mask  # Who is sending, x or kv
        mask = merge_masks(s_mask, attn_mask, q.shape)
        dropout = self.dropout if self.training else 0.0

        a_out11 = torch_attn(q1, k1, v1, mask, dropout, self.attn_type)
        a_out12 = torch_attn(q1, k1, v2, mask, dropout, self.attn_type)
        a_out1 = torch.cat([a_out11, a_out12], dim=-1)

        a_out21 = torch_attn(q2, k2, v1, mask, dropout, self.attn_type)
        a_out22 = torch_attn(q2, k2, v2, mask, dropout, self.attn_type)
        a_out2 = torch.cat([a_out21, a_out22], dim=-1)

        a_out = self._weight_by_lambda(a_out1, a_out2)

        # recombine heads
        a_out = a_out.transpose(1, 2).contiguous().view(B, S, D)

        # mix with final linear layer
        return self.out_proj(a_out)

    def forward(
        self,
        x: Tensor,
        kv: Tensor | None = None,
        mask: BoolTensor | None = None,
        kv_mask: BoolTensor | None = None,
        attn_mask: BoolTensor | None = None,
        culens: Tensor | None = None,
        maxlen: int | None = None,
    ) -> Tensor:
        """Attention forward pass, dispatches to the appropriate backend.

        Parameters
        ----------
        x : Tensor
            The pointcloud of shape (batch, x_len, dim).
        kv : Tensor
            Optional second pointcloud for cross-attn with shape (batch, kv_len, dim).
        mask : BoolTensor, optional
            Mask for the pointcloud x, by default None.
        kv_mask : BoolTensor, optional
            Mask the kv pointcloud, by default None.
        attn_mask : BoolTensor, optional
            Full attention mask, by default None.
        culens : Tensor, optional
            Cumulative lengths of the sequences in x, by default None.
            Only used for the flash-varlen backend.
        maxlen : int, optional
            Maximum length of a sequence in the x, by default None.
            Only used for the flash-varlen backend.

        Returns
        -------
        Tensor
            Output of shape (batch, x_len, dim).
        """
        if self.attn_type == "flash-varlen":
            assert kv is None, "flash-varlen only supports self attention!"
            assert attn_mask is None, "flash-varlen does not support attention masks!"
            assert culens is not None, "flash-varlen requires culens!"
            assert maxlen is not None, "flash-varlen requires maxlen!"
            if self.diff_attention:
                return self._flash_diff_forward(x, culens, maxlen)
            return self._flash_forward(x, culens, maxlen)

        # Otherwise perform standard attention
        if self.diff_attention:
            return self._torch_diff_forward(x, kv, mask, kv_mask, attn_mask)
        return self._torch_forward(x, kv, mask, kv_mask, attn_mask)


class GLU(nn.Module):
    def __init__(
        self,
        embed_dim: int,
        hidden_dim: int | None = None,
        activation: str = "SiLU",
        dropout: float = 0.0,
        bias: bool = True,
        gated: bool = False,
    ):
        """Dense update with gated linear unit.

        See [2002.05202](https://arxiv.org/abs/2002.05202).

        Parameters
        ----------
        embed_dim : int
            Dimension of the input and output.
        hidden_dim : int | None, optional
            Dimension of the hidden layer. If None, defaults to embed_dim * 2.
        activation : str, optional
            Activation function.
        dropout : float, optional
            Dropout rate.
        bias : bool, optional
            Whether to include bias in the linear layers.
        gated : bool, optional
            Whether to gate the output of the hidden layer.
        """
        super().__init__()

        if hidden_dim is None:
            hidden_dim = embed_dim * 2

        self.gated = gated
        self.embed_dim = embed_dim
        self.in_proj = nn.Linear(embed_dim, hidden_dim + hidden_dim * gated, bias=bias)
        self.out_proj = nn.Linear(hidden_dim, embed_dim, bias=bias)
        self.drop = nn.Dropout(dropout)
        self.activation = getattr(nn, activation)()

    def forward(self, x: Tensor) -> Tensor:
        x = self.in_proj(x)
        if self.gated:
            x1, x2 = x.chunk(2, dim=-1)
            x = self.activation(x1) * x2
        else:
            x = self.activation(x)
        x = self.drop(x)
        return self.out_proj(x)


class LayerScale(nn.Module):
    """Applies the LayerScale operation from the Cait vision transformer.

    Effective at improving stability and speed of deep transformers.
    Now the standard for vision transformers
    https://arxiv.org/abs/2103.17239
    """

    def __init__(self, dim: int, init_value: float = 1e-3) -> None:
        super().__init__()
        self.gamma = nn.Parameter(init_value * torch.ones(dim))

    def forward(self, x: Tensor) -> Tensor:
        return x * self.gamma


class DropPath(nn.Module):
    """Drop paths for a stochastic depth neural network.

    Used for regularisation when applied to the main path of a residual block.
    """

    def __init__(self, drop_prob: float = 0.0):
        super().__init__()
        self.drop_prob = drop_prob

    def forward(self, x: Tensor) -> Tensor:
        if self.drop_prob == 0.0 or not self.training:
            return x
        keep_prob = 1 - self.drop_prob
        shape = (x.shape[0],) + (1,) * (x.ndim - 1)
        random_tensor = keep_prob + torch.rand(shape, dtype=x.dtype, device=x.device)
        random_tensor.floor_()  # binarize
        return x.div(keep_prob) * random_tensor


class NormResidual(nn.Module):
    """Wraps a module with norm with a residual connection.

    Optionally also applies:
    - LayerScale
    - DropPath (Stochastic Depth)

    Can represent the most common transformer patterns:
    - PostNorm: x = norm(x + drop(scale * fn(x)))
    - PreNorm: x = x + drop(scale * fn(norm(x)))
    - NoNorm x = x + drop(scale * fn(x))
    """

    def __init__(
        self,
        fn: nn.Module,
        norm: str = "LayerNorm",
        ls_init: float | None = None,
        drop_path: float = 0.0,
        embed_dim: int = 0,
        norm_type: str = "pre",
    ) -> None:
        """Parameters
        ----------
        fn : nn.Module
            The module to wrap. Must be non-resizing.
        norm : str, optional
            The normalization method, by default "LayerNorm".
        ls_init : float | None, optional
            The initial value for the layerscale, by default 1e-3.
            If None, then no layerscale is applied.
        drop_path : float, optional
            The drop path rate, by default 0.0.
        embed_dim : int
            The dimension of the input and output.
            If zero we will try get it from the fn's own embed_dim attribute.
        norm_type : str, optional
            The type of normalization, by default "pre". Can be ['pre', 'post', 'none'].
        """
        super().__init__()
        self.norm_type = norm_type
        dim = embed_dim or fn.embed_dim
        assert dim > 0, "Could not determine embed_dim from fn"
        self.fn = fn
        if self.norm_type != "none":
            self.norm = getattr(layernorms, norm)(dim)
        self.ls = LayerScale(dim, ls_init) if ls_init is not None else nn.Identity()
        self.drop_path = DropPath(drop_path) if drop_path else nn.Identity()

    def forward(self, x: Tensor, *args, **kwargs) -> Tensor:
        if self.norm_type == "pre":
            return x + self.drop_path(self.ls(self.fn(self.norm(x), *args, **kwargs)))
        if self.norm_type == "post":
            return self.norm(x + self.drop_path(self.ls(self.fn(x, *args, **kwargs))))
        return x + self.drop_path(self.ls(self.fn(x, *args, **kwargs)))


class EncoderLayer(nn.Module):
    def __init__(
        self,
        embed_dim: int,
        norm: str = "LayerNorm",
        ls_init: float | None = None,
        drop_path: float = 0.0,
        depth: int = 1,
        dense_kwargs: dict | None = None,
        attn_kwargs: dict | None = None,
        norm_type: str = "pre",
    ) -> None:
        """Encoder layer consisting of a self-attention and a feed-forward layer.

        Parameters
        ----------
        embed_dim : int
            Dimension of the embeddings at each layer.
        norm : str, optional
            Normalization style, by default "LayerNorm".
        drop_path : float, optional
            Drop path rate, by default 0.0.
        ls_init : float | None, optional
            Initial value for the layerscale, by default 1e-3.
        depth : int, optional
            The depth of the layer, by default 1.
        dense_kwargs : dict | None, optional
            Keyword arguments for [salt.models.transformer_v2.GLU][salt.models.transformer_v2.GLU].
        attn_kwargs : dict | None, optional
            Keyword arguments for
            [salt.models.transformer_v2.Attention][salt.models.transformer_v2.Attention].
        norm_type : str, optional
            Normalization type, can be ['pre', 'post', 'hybrid'], by default 'pre'.
        """
        super().__init__()

        # Safe defaults
        if attn_kwargs is None:
            attn_kwargs = {}
        if dense_kwargs is None:
            dense_kwargs = {}

        # Attributes
        self.embed_dim = embed_dim
        self.norm_type = norm_type
        if norm_type == "hybrid":
            attn_kwargs["do_qk_norm"] = True
            attn_kwargs["do_v_norm"] = True
            residual_norm_type = "pre" if depth == 0 else "none"
            self.norm = (
                nn.Identity(embed_dim) if depth == 0 else getattr(layernorms, norm)(embed_dim)
            )
        else:
            residual_norm_type = norm_type

        # Submodules
        residual = partial(
            NormResidual,
            norm=norm,
            ls_init=ls_init,
            drop_path=drop_path,
            norm_type=residual_norm_type,
        )
        self.attn = residual(Attention(embed_dim, depth=depth, **attn_kwargs))
        self.dense = residual(GLU(embed_dim, **dense_kwargs))

    def forward(self, x: Tensor, **kwargs) -> Tensor:
        x = self.attn(x, **kwargs)
        if self.norm_type == "hybrid":
            return self.dense(self.norm(x))
        return self.dense(x)


class DecoderLayer(nn.Module):
    def __init__(
        self,
        embed_dim: int,
        norm: str = "LayerNorm",
        ls_init: float | None = 1e-3,
        drop_path: float = 0.0,
        depth: int = 1,
        dense_kwargs: dict | None = None,
        attn_kwargs: dict | None = None,
        norm_type: str = "pre",
    ):
        super().__init__()

        # Safe defaults
        if attn_kwargs is None:
            attn_kwargs = {}
        if dense_kwargs is None:
            dense_kwargs = {}

        # Attributes
        self.embed_dim = embed_dim

        # Submodules
        residual = partial(
            NormResidual, norm=norm, ls_init=ls_init, drop_path=drop_path, norm_type=norm_type
        )
        self.self_attn = residual(Attention(embed_dim=embed_dim, depth=depth, **attn_kwargs))
        self.cross_attn = residual(Attention(embed_dim=embed_dim, depth=depth, **attn_kwargs))
        self.dense = residual(GLU(embed_dim, **dense_kwargs))

    def forward(
        self,
        x: Tensor,
        *,  # Indicates that kv is required
        kv: Tensor,
        mask: Tensor | None = None,
        kv_mask: Tensor | None = None,
    ) -> Tensor:
        x = self.self_attn(x, kv_mask=mask)
        x = self.cross_attn(x, kv=kv, kv_mask=kv_mask)
        return self.dense(x)


class TransformerV2(nn.Module):
    def __init__(
        self,
        num_layers: int,
        embed_dim: int,
        out_dim: int | None = None,
        norm: str = "LayerNorm",
        attn_type: str = "torch-math",
        do_final_norm: bool = True,
        num_registers: int = 1,
        drop_registers: bool = False,
        **kwargs,
    ) -> None:
        """Transformer model consisting of a stack of Transformer encoder layers.

        Parameters
        ----------
        num_layers : int
            Number of layers.
        embed_dim : int
            Dimension of the embeddings at each layer.
        out_dim : int | None, optional
            Optionally project the output to a different dimension.
        norm : str, optional
            Normalization style, by default "LayerNorm".
        attn_type : str, optional
            The backend for the attention mechanism, by default "torch-flash".
            Provided here because the varlen backend requires pre/post processing.
        do_final_norm : bool, optional
            Whether to apply a final normalization layer, by default True.
        num_registers : int, optional
            The number of registers to add to the END of the input sequence.
            Registers are randomly initialised tokens of the same dimension as
            any other inputs after initialiser networks. See 2309.16588.
        drop_registers : bool, optional
            If to drop the registers from the outputs
        kwargs : dict
            Keyword arguments for [salt.models.transformer_v2.EncoderLayer].
        """
        super().__init__()

        # Check the inputs
        if num_registers < 1:
            raise ValueError(
                "Some jets have no tracks, which causes NaNs in the attention scores. ",
                "To avoid this, set num_registers to at least 1",
            )

        # Attributes
        self.num_layers = num_layers
        self.embed_dim = embed_dim
        self.out_dim = out_dim or embed_dim
        self.do_final_norm = do_final_norm
        self.do_out_proj = out_dim is not None
        self.attn_type = attn_type
        self.num_registers = num_registers
        self.drop_registers = drop_registers

        # Submodules
        kwargs["attn_kwargs"]["attn_type"] = self.attn_type
        self.layers = torch.nn.ModuleList([
            EncoderLayer(embed_dim=embed_dim, norm=norm, depth=depth, **kwargs)
            for depth in range(num_layers)
        ])

        # Check and set the attention type
        assert self.attn_type in ATTN_TYPES, "Invalid attention type!"
        self.set_backend(self.attn_type)

        # Optional submodules
        if self.do_out_proj:
            self.out_proj = nn.Linear(self.embed_dim, out_dim)
        if self.do_final_norm:
            self.out_norm = getattr(layernorms, norm)(self.out_dim)
        if self.num_registers:
            self.registers = nn.Parameter(
                torch.normal(torch.zeros((self.num_registers, self.embed_dim)), std=1e-4)
            )
            self.register_buffer("register_mask", torch.zeros(num_registers, dtype=torch.bool))
        self.featurewise = nn.ModuleList()

    def set_backend(self, attn_type: str):
        self.attn_type = attn_type
        for layer in self.layers:
            self.attn_type = layer.attn.fn.set_backend(self.attn_type)

    def forward(
        self,
        x: Tensor,
        pad_mask: BoolTensor,
        inputs: Tensors | None = None,
        **kwargs,
    ) -> Tensor:
        # Add the registers to the sequence and the mask
        if self.num_registers:
            x, pad_mask = self._add_registers(x, pad_mask)

        # Combine the input sequences if they are dictionaries (don't overwrite pad_mask)
        if isinstance(x, dict):
            x = torch.cat(list(x.values()), dim=1)
        mask = torch.cat(list(pad_mask.values()), dim=1) if isinstance(pad_mask, dict) else pad_mask

        # If using the varlen backend, pack the sequence and store the cumulative lengths
        if self.attn_type == "flash-varlen":
            x, kwargs["culens"], kwargs["maxlen"] = undo_padding(x, mask)

        # Run through the main transformer encoder layers
        for i, layer in enumerate(self.layers):
            if len(self.featurewise) > 0:
                x = self.featurewise[i](inputs, x)
            x = layer(x, mask=mask, **kwargs)

        # Run through the optional layers
        if self.do_out_proj:
            x = self.out_proj(x)
        if self.do_final_norm:
            x = self.out_norm(x)

        # If using the varlen backend, unpack the sequence
        if self.attn_type == "flash-varlen":
            x = redo_padding(x, mask)

        # Optionally drop the registers from the output
        if self.drop_registers:
            x = x[:, : -self.num_registers]
            if isinstance(pad_mask, dict):
                del pad_mask["REGISTERS"]
            elif isinstance(pad_mask, Tensor):
                pad_mask = pad_mask[:, : -self.num_registers]

        return x, pad_mask

    def _add_registers(self, x: Tensor | dict, pad_mask: BoolTensor | dict | None) -> tuple:
        """Add the learnable registers to the end of the input sequence."""
        # Get the batch size and expand the registers to match
        B = next(iter(x.values())).size(0) if isinstance(x, dict) else x.size(0)

        # Add as a key or concatenate at the end
        reg = self.registers.expand(B, -1, -1)
        if isinstance(x, dict):
            x["REGISTERS"] = reg
        else:
            x = torch.cat([x, reg], dim=1)

        # Also include a mask for the registers
        if pad_mask is not None:
            reg_mask = self.register_mask.expand(B, -1)
            if isinstance(pad_mask, dict):
                pad_mask["REGISTERS"] = reg_mask
            else:
                pad_mask = torch.cat([pad_mask, reg_mask], dim=-1)

        return x, pad_mask



import copy

import numpy as np
import pytest
import torch
from torch import nn

from salt.models import (
    Dense,
    GATv2Attention,
    GlobalAttentionPooling,
    MultiheadAttention,
    ScaledDotProductAttention,
    TensorCrossAttentionPooling,
    TransformerCrossAttentionEncoder,
    TransformerEncoder,
)
from salt.utils.inputs import get_random_mask


def test_dense() -> None:
    net = Dense(10, 10, [10, 10], activation="ReLU")
    net(torch.rand(10))


def test_dense_context() -> None:
    net = Dense(10, 10, [10, 10], activation="ReLU", context_size=4)
    net(torch.rand(10), torch.rand(4))


def test_dense_context_broadcast() -> None:
    net = Dense(10, 10, [10, 10], activation="ReLU", context_size=4)
    net(torch.rand(1, 10, 10), torch.rand(1, 4))


@pytest.mark.parametrize("pooling", [GlobalAttentionPooling, TensorCrossAttentionPooling])
def test_pooling(pooling) -> None:
    if pooling != GlobalAttentionPooling:
        net = pooling(10, 1, {"num_heads": 1, "attention": ScaledDotProductAttention()})
    else:
        net = pooling(10)

    x = {"emb": torch.rand(1, 5, 10)}
    out = net(x)

    x = {"emb": torch.cat([x["emb"], torch.zeros((1, 1, x["emb"].shape[2]))], dim=1)}
    mask = get_random_mask(1, 6, p_valid=1)
    mask[:, -1] = True
    mask = {"mask": mask}
    out_with_mask = net(x, pad_mask=mask)
    assert torch.all(out == out_with_mask)


def test_transformer() -> None:
    net = TransformerEncoder(
        embed_dim=10,
        num_layers=2,
        mha_config={
            "num_heads": 2,
            "attention": ScaledDotProductAttention(),
        },
        dense_config={
            "activation": "ReLU",
            "hidden_layers": [10],
        },
    )
    # basic test
    net(torch.rand(10, 10, 10))

    # test zero track case
    assert torch.all(net(torch.rand(10, 0, 10)) == torch.empty((10, 0, 10)))

    # test fully padded case
    out = net(torch.rand(1, 10, 10), pad_mask=get_random_mask(1, 10, p_valid=0))
    assert not torch.isnan(out).any()
    """
    disabling for now as it's failing in the CI
    # test that adding a padded track does not change the output
    tracks = torch.rand(1, 10, 10)
    mask = get_random_mask(1, 10, p_valid=1)
    out = net(tracks, pad_mask=mask)
    tracks = torch.cat([tracks, torch.zeros((1, 1, tracks.shape[2]))], dim=1)
    mask = torch.zeros(tracks.shape[:-1]).bool()
    mask[:, -1] = True
    assert torch.all(out.data == out_with_pad.data)
    out_with_pad = net(tracks, pad_mask=mask)[:, :-1]
    """


def test_transformer_cross_attention_encoder() -> None:
    net = TransformerCrossAttentionEncoder(
        input_names=["type1", "type2"],
        embed_dim=10,
        num_layers=2,
        mha_config={
            "num_heads": 2,
            "attention": ScaledDotProductAttention(),
        },
        sa_dense_config={
            "activation": "ReLU",
            "hidden_layers": [10],
        },
    )
    # Basic Functionality Test
    x = {
        "type1": torch.rand(10, 10, 10),
        "type2": torch.rand(10, 10, 10),
    }
    mask = {
        "type1": get_random_mask(10, 10, p_valid=1),
        "type2": get_random_mask(10, 10, p_valid=1),
    }
    net(x, mask)

    # Zero Input Case Test
    x["type1"] = torch.rand(10, 0, 10)
    mask["type1"] = torch.empty((10, 0), dtype=bool)
    assert torch.all(net(x, mask)["type1"] == torch.empty((10, 0, 10)))

    # Padded Case Test
    x["type1"] = torch.rand(1, 10, 10)
    x["type2"] = torch.rand(1, 10, 10)
    mask["type1"] = get_random_mask(1, 10, p_valid=0)
    mask["type2"] = get_random_mask(1, 10, p_valid=1)
    out = net(x, mask)
    assert not torch.isnan(out["type1"]).any()

    # Padding Invariance Test
    del x
    x = {
        "type1": torch.rand(1, 10, 10),
        "type2": torch.rand(1, 10, 10),
    }
    extended_x = copy.deepcopy(x)
    del extended_x["type1"]
    extended_x.update({
        "type1": torch.cat([copy.deepcopy(x["type1"]), torch.zeros((1, 1, 10))], dim=1)
    })
    mask["type1"] = get_random_mask(1, 10, p_valid=1)
    out = net(x, mask)
    mask["type1"] = torch.zeros(extended_x["type1"].shape[:-1]).bool()
    mask["type1"][:, -1] = True
    out_with_pad = net(extended_x, mask)["type1"][:, :-1]
    torch.testing.assert_close(out["type1"], out_with_pad)


def test_mha_allvalid_mask() -> None:
    n_batch = 2
    n_trk = 3
    n_head = 1
    n_dim = 5

    net = MultiheadAttention(
        embed_dim=n_dim,
        num_heads=n_head,
        attention=ScaledDotProductAttention(),
    )

    q = k = v = torch.rand((n_batch, n_trk, n_dim))
    valid_mask = torch.zeros((n_batch, n_trk), dtype=torch.bool)
    out_no_mask = net(q, k, v, q_mask=None, kv_mask=None)
    out_mask = net(q, k, v, q_mask=valid_mask, kv_mask=valid_mask)
    assert torch.all(out_no_mask == out_mask)


def test_gatv2():
    n_batch = 10
    n_trk = 10
    n_head = 4
    n_dim = 16
    head_dim = n_dim // n_head

    net = MultiheadAttention(
        attention=GATv2Attention(n_head, head_dim, activation="SiLU"),
        embed_dim=n_dim,
        num_heads=n_head,
    )

    q = k = v = torch.rand((n_batch, n_trk, n_dim))
    valid_mask = torch.zeros((n_batch, n_trk), dtype=torch.bool)
    out_no_mask = net(q, k, v, q_mask=None, kv_mask=None)
    out_mask = net(q, k, v, q_mask=valid_mask, kv_mask=valid_mask)
    assert torch.all(out_no_mask == out_mask)

    q_mask = kv_mask = get_random_mask(n_batch, n_trk)
    net(q, k, v, q_mask=q_mask, kv_mask=kv_mask)


def test_mha_qkv_different_dims():
    n_batch = 3
    n_trk = 5
    n_head = 2
    k_dim = 7
    q_dim = 8
    v_dim = 9

    net = MultiheadAttention(
        num_heads=n_head,
        embed_dim=q_dim,
        k_dim=k_dim,
        v_dim=v_dim,
        attention=ScaledDotProductAttention(),
    )

    q = torch.rand((n_batch, n_trk, q_dim))
    k = torch.rand((n_batch, n_trk + 1, k_dim))
    v = torch.rand((n_batch, n_trk + 1, v_dim))

    q_mask = get_random_mask(n_batch, n_trk)
    kv_mask = get_random_mask(n_batch, n_trk + 1)
    net(q, k, v, q_mask=q_mask, kv_mask=kv_mask)


@pytest.mark.parametrize("frac_pad", [0.0, 0.5, 1.0])
def test_mha_vs_torch(frac_pad):
    torch.manual_seed(0)
    n_batch = 4
    n_trk = 10
    n_head = 1
    n_dim = 8
    frac_pad = 0.2

    t_net, s_net = get_pytorch_salt_mha(n_dim, n_head)

    mask = torch.zeros((n_batch, n_trk), dtype=torch.bool)  # all valid
    mask[..., : int(frac_pad * n_trk)] = True  # set some padding

    x = torch.rand((n_batch, n_trk, n_dim))
    torch_out = t_net(x, x, x, key_padding_mask=mask)[0]
    out = s_net(x, x, x, kv_mask=mask)

    np.testing.assert_allclose(
        torch_out.detach().numpy(), out.detach().numpy(), rtol=1e-6, atol=1e-6
    )


def get_pytorch_salt_mha(n_dim, n_head):
    """Get a pytorch and salt MHA layer with equivalent weights."""
    t_net = nn.MultiheadAttention(
        embed_dim=n_dim, num_heads=n_head, batch_first=True, bias=True, add_zero_attn=True
    )

    s_net = MultiheadAttention(
        attention=ScaledDotProductAttention(), embed_dim=n_dim, num_heads=n_head
    )

    weights = torch.rand((3 * n_dim, n_dim))
    bias = torch.rand(3 * n_dim)
    t_net.in_proj_weight = nn.Parameter(weights)
    t_net.in_proj_bias = nn.Parameter(bias)

    wq, wk, wv = weights.chunk(3)
    bq, bk, bv = bias.chunk(3)
    s_net.linear_q.weight = nn.Parameter(wq)
    s_net.linear_k.weight = nn.Parameter(wk)
    s_net.linear_v.weight = nn.Parameter(wv)
    s_net.linear_q.bias = nn.Parameter(bq)
    s_net.linear_k.bias = nn.Parameter(bk)
    s_net.linear_v.bias = nn.Parameter(bv)

    s_net.linear_out.weight = t_net.out_proj.weight
    s_net.linear_out.bias = t_net.out_proj.bias

    return t_net, s_net



import torch
from torch import Tensor, nn
from torch.nn import ModuleList


class R21Xbb(nn.Module):
    def __init__(self, tasks: ModuleList):
        super().__init__()
        self.tasks = tasks
        self.init_nets: list = []

    def forward(self, inputs: dict, pad_masks=None, labels: dict | None = None):
        _ = pad_masks
        x = inputs["track"]
        x = torch.flatten(x, start_dim=1)
        preds, loss = self.run_tasks(x, labels)
        return preds, loss

    def run_tasks(
        self,
        x: Tensor,
        labels: dict | None = None,
    ):
        preds = {}
        loss = {}
        for task in self.tasks:
            if task.input_name == self.tasks[0].global_object:
                task_input = x
            else:
                print("WARNING: R21Xbb is not configured for input_name other than jet")
                print("WARNING: R21Xbb shall go on jet tagging")
                task_input = x
            task_preds, task_loss = task(task_input, labels, None, context=None)
            preds[task.name] = task_preds
            loss[task.name] = task_loss

        return preds, loss



"""Implements the loss of MaskFormer, utilising a Hungarian matcher.

Based on
- https://github.com/facebookresearch/MaskFormer
- https://github.com/facebookresearch/Mask2Former
"""

import torch
import torch.nn.functional as F
from torch import Tensor, nn

from salt.models.matcher import HungarianMatcher


@torch.jit.script
def dice_loss(inputs: Tensor, labels: Tensor):
    """Compute the DICE loss, similar to generalized IOU for masks.

    Parameters
    ----------
    inputs : Tensor
        The predictions for each example.
    labels : Tensor
        A float tensor with the same shape as inputs. Stores the binary classification label
        for each element in inputs (0 for the negative class and 1 for the positive class).

    Returns
    -------
    Tensor
        Single-element loss tensor
    """
    inputs = inputs.sigmoid()
    numerator = 2 * (inputs * labels).sum(-1)
    denominator = inputs.sum(-1) + labels.sum(-1)
    loss = 1 - (numerator + 1) / (denominator + 1)
    return loss.sum() / len(inputs)


@torch.jit.script
def mask_ce_loss(inputs: Tensor, labels: Tensor):
    """Computes cross entropy loss for masks.

    Parameters
    ----------
    inputs: Tensor
            A float tensor of arbitrary shape representing the predictions for each example.
    labels: Tensor
        A float tensor with the same shape as inputs. Stores the binary classification label
        for each element in inputs (0 for the negative class and 1 for the positive class).

    Returns
    -------
    Tensor
        Single-element loss tensor
    """
    loss = F.binary_cross_entropy_with_logits(inputs, labels, reduction="none")
    # find the mean loss for each mask
    loss = loss.mean(1)

    # take the average over all masks
    return loss.sum() / len(inputs)


@torch.jit.script
def sigmoid_focal_loss(inputs: Tensor, targets: Tensor, alpha: float = -1, gamma: float = 2):
    """Loss used in RetinaNet for dense detection: https://arxiv.org/abs/1708.02002.

    Parameters
    ----------
    inputs: Tensor
        A float tensor of arbitrary shape representing the predictions for each example.
    targets: Tensor
        A float tensor with the same shape as inputs. Stores the binary classification label for
        each element in inputs (0 for the negative class and 1 for the positive class).
    alpha: float, optional
        Weighting factor in range (0,1) to balance positive vs negative examples.
        Default = -1 (no weighting).
    gamma: float, optional
        Exponent of the modulating factor (1 - p_t) to balance easy vs hard examples. Default is 2

    Returns
    -------
    Tensor
        Single-element loss tensor
    """
    prob = inputs.sigmoid()
    ce_loss = F.binary_cross_entropy_with_logits(inputs, targets, reduction="none")
    p_t = prob * targets + (1 - prob) * (1 - targets)
    loss = ce_loss * ((1 - p_t) ** gamma)

    if alpha >= 0:
        alpha_t = alpha * targets + (1 - alpha) * (1 - targets)
        loss = alpha_t * loss

    return loss.mean(1).sum() / len(inputs)


class MaskFormerLoss(nn.Module):
    """Compute the loss of MaskFormer, based on DETR.

    The process happens in two steps:
        1) we compute hungarian assignment between ground truth boxes and the preds of the model
        2) we supervise each pair of matched ground-truth / prediction (supervise class and box).
    """

    def __init__(
        self,
        num_classes: int,
        num_objects: int,
        loss_weights: dict,
        matcher_weights: dict | None = None,
        null_class_weight: float = 0.5,
        losses: list[str] | None = None,
        # tasks: nn.ModuleList = None,
    ):
        """Create the criterion.

        Parameters
        ----------
        num_classes: int
            number of object categories, omitting the special no-object category
        num_objects: int
            number of objects to detect
        loss_weights: dict
            dict containing as key the names of the losses and as values their relative weight
        matcher_weights: dict, optional
            same as loss_weights but for the matching cost
        null_class_weight: float, optional
            relative classification weight applied to the no-object category
        losses: list, optional
            list of all the losses to be applied. See get_loss for list of available losses
        """
        super().__init__()
        self.num_classes = num_classes
        self.null_class_weight = null_class_weight
        assert self.num_classes > 0
        if self.num_classes == 1:
            empty_weight = torch.tensor([self.null_class_weight])
        else:
            empty_weight = torch.ones(self.num_classes + 1)
            empty_weight[-1] = self.null_class_weight
        self.register_buffer("empty_weight", empty_weight)
        self.loss_weights = loss_weights
        if matcher_weights is None:
            matcher_weights = loss_weights
        self.losses = losses if losses is not None else ["labels", "masks"]

        self.matcher = HungarianMatcher(
            num_classes=num_classes,
            num_objects=num_objects,
            loss_weights=matcher_weights,
        )

    def loss_labels(self, preds, labels):
        """Classification loss (NLL)
        labels dicts must contain the key "labels" containing a tensor of dim [nb_target_boxes].
        """
        # use the new indices to calculate the loss
        # process full inidices
        flav_pred_logits = preds["class_logits"].flatten(0, 1)
        flavour_labels = labels["object_class"].flatten(0, 1)
        if flav_pred_logits.shape[1] == 1:
            loss = F.binary_cross_entropy_with_logits(
                flav_pred_logits.squeeze(), flavour_labels.float(), pos_weight=self.empty_weight
            )
        else:
            loss = F.cross_entropy(flav_pred_logits, flavour_labels, self.empty_weight)
        return {"object_class_ce": loss}

    def loss_masks(self, preds, labels):
        """Compute the losses related to the masks: the focal loss and the dice loss.
        labels dicts must contain the key "masks" containing a tensor of dim
            [nb_target_boxes, h, w].
        """
        # select valid masks via flavour label
        valid_idx = labels["object_class"] != self.num_classes
        target_masks = labels["masks"][valid_idx].float()
        pred_masks = preds["masks"][valid_idx]
        # pred_masks = preds["masks"][valid_idx][:, : target_masks.shape[-1]] # for padding

        # compute losses on valid masks
        losses = {}
        if self.loss_weights.get("mask_dice"):
            losses["mask_dice"] = dice_loss(pred_masks, target_masks)
        if self.loss_weights.get("mask_focal"):
            losses["mask_focal"] = sigmoid_focal_loss(pred_masks, target_masks)
        if self.loss_weights.get("mask_ce"):
            losses["mask_ce"] = mask_ce_loss(pred_masks, target_masks)
        return losses

    def get_loss(self, loss, preds, labels):
        loss_map = {"labels": self.loss_labels, "masks": self.loss_masks}
        assert loss in loss_map, f"do you really want to compute {loss} loss?"
        return self.weight_loss(loss_map[loss](preds["objects"], labels["objects"]))

    def weight_loss(self, losses: dict):
        """Apply the loss weights to the loss dict."""
        for k in list(losses.keys()):
            losses[k] *= self.loss_weights[k]
        return losses

    def forward(self, preds, tasks, labels):
        """Calculate the maskformer loss via optimal assignment."""
        losses = {}

        # loop over intermediate outputs and compute losses
        if "intermediate_outputs" in preds:
            for i, aux_pred in enumerate(preds["intermediate_outputs"]):
                for task in tasks:  # add regression prediction for cost
                    if task.input_name == "objects":
                        aux_pred.update(task(aux_pred, labels))

                aux_idx = self.matcher(aux_pred, labels)
                for k, v in aux_pred.items():
                    if k in {"x", "embed_xs", "global_rep"}:
                        continue
                    aux_pred[k] = v[aux_idx]

                for loss in self.losses:
                    l_dict = self.get_loss(loss, aux_pred, labels)
                    l_dict = {k + f"_layer{i}": v for k, v in l_dict.items()}
                    losses.update(l_dict)

        # run tasks on the objects (e.g. regression) as these should be included in the matching
        for task in tasks:
            if task.input_name == "objects":
                # Get the scaled targets for this task and store them in the labels dict
                # for the matcher to use
                task_targets = task.get_targets(labels)

                task_pred, _ = task(preds["objects"]["embed"], labels)

                preds["objects"].update({task.name: task_pred})
                labels["objects"][task.name] = task_targets

        # get the optimal assignment of the predictions to the labels
        idx = self.matcher(preds["objects"], labels["objects"])

        # warning: don't put this into a function or comprehension
        for k, v in preds["objects"].items():
            if k in {"x", "embed"}:
                continue

            if k != "intermediate_outputs":  # don't permute input reps
                preds["objects"][k] = v[idx]

        # compute the requested losses
        for loss in self.losses:
            losses.update(self.get_loss(loss, preds, labels))

        return preds, labels, losses



CLASS_NAMES = {
    "ftagTruthOriginLabel": [
        "Pileup",
        "Fake",
        "Primary",
        "FromB",
        "FromBC",
        "FromC",
        "FromTau",
        "OtherSecondary",
    ],
    "ftagTruthTypeLabel": [
        "NoTruth",
        "Other",
        "Pion",
        "Kaon",
        "Electron",
        "Muon",
    ],
    "ftagTruthSourceLabel": [
        "NoTruth",
        "NotSecondary",
        "HadronicInteraction",
        "StrangeMesonDecay",
        "StrangeBaryonDecay",
        "GammaConversion",
        "Other",
    ],
}



import torch
from torch import Tensor


def symmetrize_edge_scores(scores: Tensor, node_numbers: Tensor):
    """Function to make edge scores symmetric.

    Output has same format as input: (edges in batch, 1). Node number
    array gives number of nodes per graph in batch as a tensor.
    """
    node_numbers = node_numbers[node_numbers > 1].long()  # remove jets without edges
    edge_numbers = (node_numbers * (node_numbers - 1)).long()

    # calculate cumulative edge numbers and offsets for symmetric index calculation
    cum_edges = torch.cumsum(edge_numbers, 0).long()
    edge_offsets = torch.cat([torch.tensor([0], device=cum_edges.device).long(), cum_edges[:-1]])

    torch.tensor_split(scores, cum_edges)

    # calculate opposite edge indices (assumes edges sorted by src then dest or vice versa)
    sym_ind = torch.cat([
        torch.arange(n - 1, n * (n - 1) ** 2 + 1, n - 1, device=node_numbers.device)
        for n in node_numbers
    ])
    sym_ind += torch.cat([
        torch.arange(0, n - 1, device=sym_ind.device).repeat(n).sort()[0] * n for n in node_numbers
    ])
    sym_ind = sym_ind % torch.cat([n.repeat(n) for n in edge_numbers])
    sym_ind += torch.cat([
        edge_offsets[i].repeat(edge_numbers[i]) for i in range(len(edge_offsets))
    ])

    edge_scores = (scores + scores[sym_ind]) / 2.0

    return torch.sigmoid(edge_scores.float())


def update_node_indices(
    scores: Tensor, node_indices: Tensor, update_indices: Tensor, node_numbers: Tensor
):
    """Run a single step of the union find algorithm.

    Takes a score matrix with shape (edges in batch, 1) and a tensor with
    the number of nodes in each graph of the batch as well as a tensor specifying
    for which graphs the algorithm has already terminated. Returns updated vertex
    index and algorithm termination tensors.
    """
    edge_offset = node_offset = 0
    for i, _nnodes in enumerate(node_numbers):
        nnodes = int(_nnodes.item())
        nedges = nnodes * (nnodes - 1)

        if nnodes <= 1:
            update_indices[i] = False

        if not update_indices[i]:
            edge_offset += nedges
            node_offset += nnodes
            continue

        node_ids = torch.arange(0, nnodes)
        edge_ids = torch.transpose(
            torch.stack((node_ids.repeat(nnodes).sort()[0], node_ids.repeat(nnodes))), 0, 1
        )
        edge_ids = edge_ids[edge_ids[:, 0] != edge_ids[:, 1]]

        max_val = torch.max(node_indices[node_offset : node_offset + nnodes]) + 1

        # create arrays with node indices for source and destination nodes of each edge
        src_ids = node_indices[node_offset : node_offset + nnodes][edge_ids[:, 0]]
        dest_ids = node_indices[node_offset : node_offset + nnodes][edge_ids[:, 1]]

        # pick out minimum between source and destination node ids, filter edges with scores < 0.5
        dest_ids[scores[edge_offset : edge_offset + nedges, 0] < 0.5] = max_val
        min_ids = torch.minimum(src_ids, dest_ids)

        # get the lowest node id for each node across all edges for each node
        node_indices_prev = torch.clone(node_indices[node_offset : node_offset + nnodes])
        node_indices[node_offset : node_offset + nnodes] = torch.min(
            torch.reshape(min_ids, (nnodes, -1)), dim=-1
        )[0]

        update_indices[i] = torch.any(
            ~torch.eq(node_indices_prev, node_indices[node_offset : node_offset + nnodes])
        )

        edge_offset += nedges
        node_offset += nnodes

    return node_indices, update_indices


def get_node_assignment(output: Tensor, mask: Tensor):
    """Run edge score symmetrization and union find.

    Wrapper function which returns reconstructed vertex indices in shape
    (nodes in batch, 1). Assumes mask of shape (batch, max_tracks).
    """
    # pad mask with additional track to avoid onnx error
    mask = torch.cat(
        [mask, torch.ones((mask.shape[0], 1), dtype=torch.bool, device=mask.device)], dim=1
    )
    node_numbers = (~mask).sum(dim=-1).long()

    # symmetrize edge scores
    scores = symmetrize_edge_scores(output, node_numbers) if output.shape[0] > 0 else output

    # update node assignments until no more changes occur
    node_indices = torch.cat([torch.arange(nnodes) for nnodes in node_numbers])
    update_indices = torch.ones_like(node_numbers, dtype=torch.bool)
    while torch.any(update_indices):
        node_indices, update_indices = update_node_indices(
            scores, node_indices, update_indices, node_numbers
        )

    return node_indices.unsqueeze(-1)


@torch.jit.script
def get_node_assignment_jit(output: Tensor, mask: Tensor):
    return get_node_assignment(output, mask)



import argparse
from datetime import datetime
from pathlib import Path

from slurm_handler import SlurmHandler

# Set up argument parser
parser = argparse.ArgumentParser(description="Submit batch jobs to Slurm.")
parser.add_argument("-c", "--config", required=True, type=Path, help="Configuration file for job.")
parser.add_argument("-t", "--tag", default="salt_job", help="Tag for job to be submitted.")
parser.add_argument("-p", "--partition", default=None, type=str, help="Partition to submit job.")
parser.add_argument(
    "-cn", "--constraint", default=None, type=str, help="Constraint on requested resources."
)
parser.add_argument("-a", "--account", default=None, type=str, help="Slurm account name.")
parser.add_argument(
    "-e",
    "--environment",
    default="conda",
    choices=["conda", "singularity", "local"],
    help="Environment for job to be submitted.",
)
parser.add_argument("-q", "--qos", default=None, type=str, help="Quality Of Service for job")
parser.add_argument("-n", "--nodes", default=1, type=int, help="Nodes to split training across")
parser.add_argument("-g", "--gpus_per_node", default=1, type=int, help="GPUs for each node")
parser.add_argument(
    "-gt",
    "--gpu_type",
    default="",
    type=str,
    help="GPU type e.g. v100, leave empty for no preference",
)
parser.add_argument("-cpt", "--cpus_per_task", default=10, type=int, help="CPUs for each task")
parser.add_argument("-m", "--memory", default="100G", type=str, help="Memory per node")
parser.add_argument("-ex", "--exclusive", action="store_true")
parser.add_argument("-ti", "--time", default=None, type=str, help="Job time limit e.g. '24:00:00'")
parser.add_argument("-f", "--force", action="store_true")
parser.add_argument(
    "-b",
    "--bind",
    nargs="+",
    help="List of binds for singularity (e.g. /path/to/upp/output:/inputs)",
)
parser.add_argument("-r", "--requeue", action="store_true")
parser.add_argument(
    "-s",
    "--signal",
    default="SIGUSR1@90",
    type=str,
    help="Signal from Slurm to trigger Lightning to prepare for requeue",
)
parser.add_argument(
    "-sls",
    "--salt_log_suffix",
    default=None,
    help="Appended to model name to create Salt log directory",
)
args, extra_args = parser.parse_known_args()

# Manually parse extra args not explicitly handled for passing to Salt and Slurm
extra_slurm_args = {}
extra_config_args = {}
for extra_arg in extra_args:
    dest_argument = extra_arg.lstrip("-")
    dest = dest_argument.split(".")[0]
    argument = ".".join(dest_argument.split(".")[1:])
    if "=" not in argument:
        name = argument
        value = None
    else:
        name, value = argument.split("=")
    if dest == "slurm":
        extra_slurm_args[name] = value
    elif dest == "config":
        extra_config_args[name] = value
    else:
        raise ValueError(f'Extra argument provided but unknown destination "{dest}"')

if args.bind and args.environment != "singularity":
    parser.error("--bind option is only allowed with --environment singularity")

# Define directories
batch_dir = Path.cwd() / "slurm"
batch_path = batch_dir / "batch"
log_path = batch_dir / "batch_logs"
for directory in [batch_path, log_path]:
    directory.mkdir(parents=True, exist_ok=True)

# Variables that need to be harmonized between Slurm and salt
nodes = args.nodes
gpus_per_node = args.gpus_per_node
cpus_per_task = args.cpus_per_task

gpu_type = args.gpu_type
gres = f"gpu:{gpu_type}:{gpus_per_node}" if gpu_type else f"gpu:{gpus_per_node}"

# Set up Slurm options
job_basedir = Path(__file__).resolve().parent.parent.parent
handler = SlurmHandler(str(batch_path), str(log_path), str(job_basedir))
handler["job-name"] = args.tag
if args.partition is not None:
    handler["partition"] = args.partition
if args.constraint is not None:
    handler["constraint"] = args.constraint
if args.account is not None:
    handler["account"] = args.account
if args.qos is not None:
    handler["qos"] = args.qos
handler["nodes"] = nodes
handler["gres"] = gres
handler["ntasks-per-node"] = gpus_per_node
if args.memory is not None:
    handler["mem"] = args.memory  # memory, 100 GiB - in MiB
if args.exclusive:
    handler["exclusive"] = None  # Exclusive access to nodes
handler["cpus-per-task"] = cpus_per_task  # Don't use this if you have exclusive access to the node
handler["export"] = "ALL"
handler["output"] = f"{log_path}/slurm-%j.out"
handler["error"] = f"{log_path}/slurm-%j.err"
if args.time is not None:
    handler["time"] = args.time  # Time limit of job, default is system specified
if args.requeue:
    handler["requeue"] = None
    handler["signal"] = args.signal
# Fill extra Slurm args
for arg, value in extra_slurm_args.items():
    handler[arg] = value


log_suffix = args.salt_log_suffix
if args.requeue and not log_suffix:
    log_suffix = datetime.now().strftime("%Y%m%d-T%H%M%S")

# Construct and submit the job command
command = "cd ${BASEDIR} && " "export OMP_NUM_THREADS=1\n"
if args.environment == "conda":
    command += (
        "source conda/bin/activate && conda activate salt\n"
        'echo "Activated environment ${CONDA_DEFAULT_ENV}"\n'
    )
elif args.environment == "singularity":
    command += "srun singularity exec -e --nv \\\n"
    command += " \\\n".join([f"--bind {b}" for b in args.bind]) + " \\\n"
    command += (
        "--home ${BASEDIR} \\\n"
        "/cvmfs/unpacked.cern.ch/gitlab-registry.cern.ch/atlas-flavor-tagging-tools/algorithms/salt:latest/ \\\n"  # noqa: E501
        'sh -c "'
    )
command += (
    "echo 'CUDA_VISIBLE_DEVICES: ${CUDA_VISIBLE_DEVICES}' &&\n"
    "cat /proc/cpuinfo | awk '/^processor/{print $3}' | tail -1 &&\n"
    "cd ${BASEDIR}/salt && pwd &&\n"
    + ("srun " if args.environment == "conda" else "")
    + f"salt fit --config {args.config.resolve()} \\\n"
    f"      --trainer.devices={gpus_per_node} \\\n"
    f"      --trainer.num_nodes={nodes} \\\n"
    f"      --data.num_workers={cpus_per_task} \\\n"
)

if args.requeue:
    command += f"      --overwrite_config \\\n" f"      --log_suffix={log_suffix} \\\n"

# Fill extra args to Salt config
for arg, value in extra_config_args.items():
    if value is not None:
        command += f"      --{arg}={value} \\\n"
    else:
        command += f"      --{arg} \\\n"

if args.force:
    command += "      --force"
if args.environment == "singularity":
    command += '"'

# handler.activate_testmode() # To inspect batch script before running
handler.send_job(command, args.tag)



from lightning import Callback, LightningModule, Trainer


class GradientLoggerCallback(Callback):
    def __init__(self, log_every_n_steps=50):
        """Callback to log model gradients during training.

        Args:
            log_every_n_steps (int): Frequency of logging gradients. Logs every `n` steps.
        """
        self.log_every_n_steps = log_every_n_steps

    def setup(self, trainer: Trainer, module: LightningModule, stage: str) -> None:
        if trainer.fast_dev_run or stage != "fit":
            return
        kwargs = {"sync_dist": len(trainer.device_ids) > 1}

        def log(metrics, stage):
            for t, loss_value in metrics.items():
                n = f"{stage}_{t}"
                module.log(n, loss_value, **kwargs)

        self.log = log

    def on_after_backward(self, trainer, pl_module):
        """Called after the backward pass in training.
        Logs the gradients of the model's parameters.
        """
        # Check if logging should happen at this step
        if trainer.global_step % self.log_every_n_steps == 0:
            total_grad_magnitude = 0.0
            total_params = 0
            for name, param in pl_module.named_parameters():
                if param.grad is not None:
                    grad_magnitude = param.grad.norm().item()
                    total_grad_magnitude += grad_magnitude
                    total_params += param.grad.numel()
                    # Log gradient statistics
                    grad_mean = param.grad.mean().item()
                    grad_std = param.grad.std().item()
                    self.log(
                        {f"grad/{name}_mean": grad_mean, f"grad/{name}_std": grad_std}, "gradients"
                    )

                else:
                    self.log({f"grad/{name}_mean": None, f"grad/{name}_std": None}, "gradients")

            # Log total gradient magnitude or average magnitude
            if total_params > 0:
                avg_grad_magnitude = total_grad_magnitude / total_params
                self.log(
                    {
                        "total_magnitude": total_grad_magnitude,
                        "average_magnitude": avg_grad_magnitude,
                    },
                    "gradient",
                )



import logging
import subprocess
from pathlib import Path
from typing import Any

logging.basicConfig(level=logging.INFO)


class CondorHandler:
    """A class to submit batch jobs to a HTCondor scheduler.

    Attributes
    ----------
    batch_path : Path
        Path where the batch config files which are created will be stored.
    log_path : Path
        Path where the batch log files will be stored.
    base_dir : Path
        Directory in which batch job will execute its command.

    Methods
    -------
    activate_testmode():
        Activate test mode: check config files in dry runs, no jobs submitted.
    deactivate_testmode():
        Deactivate test mode, enable submitting jobs.
    send_job(command: str, tag: str = "htcondor_job"):
        Submit job by creating config files (bash file and HTCondor submission file)
        and executing condor_submit.
    """

    def __init__(self, batch_path: str, log_path: str, basedir: str) -> None:
        self.batch_path = Path(batch_path)
        self.log_path = Path(log_path)
        self.base_dir = Path(basedir) if basedir else Path.cwd()
        self._tag = "salt_job"
        # Keywords to be used in HTCondor configuration file
        self._condor_options_dict = {
            "universe": "Universe",
            "jobflavour": "+JobFlavour",
            "project": "+MyProject",
            "runtime": "+RequestRuntime",
            "memory": "Request_Memory",
            "cpu": "Request_CPUs",
            "gpu": "Request_GPUs",
            "requirements": "Requirements",
            "container": "+MySingularityImage",
            "containerargs": "+MySingularityArgs",
        }
        self._condor_options: dict[str, Any] = {}
        self._test_mode = False

    def activate_testmode(self) -> None:
        logging.debug("Activated test mode: not submitting any jobs.")
        self._test_mode = True

    def deactivate_testmode(self) -> None:
        logging.debug("Deactivated test mode: submitting jobs.")
        self._test_mode = False

    def send_job(self, command: str, tag: str = "salt_job") -> None:
        self._tag = tag
        bashfile = self._make_bash_file(command)
        jobfile = self._make_job_file(bashfile)
        if self._test_mode:
            logging.debug(f"Created job file {jobfile}")
        else:
            subprocess.call(f"condor_submit {jobfile}", shell=True)

    def __setitem__(self, key: str, value: Any) -> None:  # noqa: ANN401
        self._condor_options[key] = value

    def _make_bash_file(self, command: str) -> Path:
        run_file = self.batch_path / f"batch_{self._tag}.sh"
        with run_file.open("w") as fr:
            fr.write(f"""#!/bin/sh
# {self._tag} batch run script
#$ -cwd
#$ -j y
#$ -l cvmfs
BASEDIR={self.base_dir}
pwd; ls -l
{command}
ls -l
""")
        run_file.chmod(0o755)
        logging.debug(f"Made run file {run_file}")
        return run_file

    def _make_job_file(self, run_file: Path) -> Path:
        batch_file = self.batch_path / f"batch_{self._tag}.job"
        with batch_file.open("w") as fs:
            for key, value in self._condor_options.items():
                if key in self._condor_options_dict:
                    fs.write(f"{self._condor_options_dict[key]}={value}\n")
            fs.write(f"""Executable          = {run_file}
Output              = {self.log_path}/stdout_{self._tag}_$(ClusterId).txt
Error               = {self.log_path}/stderr_{self._tag}_$(ClusterId).txt
Log                 = {self.log_path}/batch_{self._tag}_$(ClusterId).log

queue
""")
        logging.debug(f"Made job file {batch_file}")
        return batch_file



import torch

from salt.models import (
    MultiheadAttention,
    ScaledDotProductAttention,
    TransformerEncoder,
    TransformerEncoderLayer,
)
from salt.utils.inputs import get_random_mask


def test_transformer_edges() -> None:
    net = TransformerEncoder(
        embed_dim=10,
        edge_embed_dim=10,
        num_layers=2,
        mha_config={
            "num_heads": 2,
            "attention": ScaledDotProductAttention(),
        },
        dense_config={
            "activation": "ReLU",
            "hidden_layers": [10],
        },
    )
    # basic test
    net(torch.rand(10, 10, 10), torch.rand(10, 10, 10, 10))

    # test zero track case
    assert torch.all(
        net(torch.rand(10, 0, 10), torch.rand(10, 0, 0, 10)) == torch.empty((10, 0, 10))
    )

    # test fully padded case
    out = net(
        torch.rand(1, 10, 10), torch.rand(1, 10, 10, 10), pad_mask=get_random_mask(1, 10, p_valid=0)
    )
    assert not torch.isnan(out).any()

    # test that adding a padded track does not change the output
    """tracks = torch.rand(1, 10, 10)
    edges = torch.rand(1, 10, 10, 10)
    mask = get_random_mask(1, 10, p_valid=1)
    out = net(tracks, edges, pad_mask=mask)
    tracks = torch.cat([tracks, torch.zeros((1, 1, tracks.shape[2]))], dim=1)
    edges = torch.cat([edges, torch.zeros((1, 1, 10, edges.shape[3]))], dim=1)
    edges = torch.cat([edges, torch.zeros((1, 11, 1, edges.shape[3]))], dim=2)
    mask = torch.zeros(tracks.shape[:-1]).bool()
    mask[:, -1] = True
    out_with_pad = net(tracks, edges, pad_mask=mask)[:, :-1]
    assert torch.all(out == out_with_pad)"""


def test_mha_edges_allvalid_mask() -> None:
    n_batch = 2
    n_trk = 3
    n_head = 1
    n_dim = 5

    net = MultiheadAttention(
        embed_dim=n_dim,
        num_heads=n_head,
        attention=ScaledDotProductAttention(),
        edge_embed_dim=n_dim,
    )

    q = k = v = torch.rand((n_batch, n_trk, n_dim))
    edges = torch.rand((n_batch, n_trk, n_trk, n_dim))
    valid_mask = torch.zeros((n_batch, n_trk), dtype=torch.bool)
    out_no_mask, _ = net(q, k, v, edges, q_mask=None, kv_mask=None)
    out_mask, _ = net(q, k, v, edges, q_mask=valid_mask, kv_mask=valid_mask)
    assert torch.all(out_no_mask == out_mask)


def test_mha_edges_qkv_different_dims():
    n_batch = 3
    n_trk = 5
    n_head = 2
    k_dim = 7
    q_dim = 8
    v_dim = 9
    edge_dim = 10

    net = MultiheadAttention(
        num_heads=n_head,
        embed_dim=q_dim,
        k_dim=k_dim,
        v_dim=v_dim,
        edge_embed_dim=edge_dim,
        attention=ScaledDotProductAttention(),
    )

    q = torch.rand((n_batch, n_trk, q_dim))
    k = torch.rand((n_batch, n_trk + 1, k_dim))
    v = torch.rand((n_batch, n_trk + 1, v_dim))
    edges = torch.rand((n_batch, n_trk, n_trk + 1, edge_dim))

    q_mask = get_random_mask(n_batch, n_trk)
    kv_mask = get_random_mask(n_batch, n_trk + 1)
    net(q, k, v, edges, q_mask=q_mask, kv_mask=kv_mask)


def test_edge_updates():
    n_batch = 2
    n_trk = 3
    n_head = 1
    n_dim = 5

    x = torch.rand((n_batch, n_trk, n_dim))
    edges = torch.rand((n_batch, n_trk, n_trk, n_dim))

    # case where edges are updated
    net = TransformerEncoderLayer(
        embed_dim=n_dim,
        mha_config={
            "num_heads": n_head,
            "attention": ScaledDotProductAttention(),
        },
        edge_embed_dim=n_dim,
        update_edges=True,
    )

    _, edges_out = net(x, edges)
    assert ~torch.all(edges == edges_out)

    # case where edges are not updated
    net = TransformerEncoderLayer(
        embed_dim=n_dim,
        mha_config={
            "num_heads": n_head,
            "attention": ScaledDotProductAttention(),
        },
        edge_embed_dim=n_dim,
        update_edges=False,
    )

    _, edges_out = net(x, edges)
    assert torch.all(edges == edges_out)



import contextlib
import json
import shutil
import socket
from contextlib import suppress
from pathlib import Path

import h5py
import lightning as L
import numpy as np
import torch
import yaml
from ftag.git_check import get_git_hash
from lightning import Callback, LightningModule, Trainer
from lightning.pytorch.cli import LightningArgumentParser, Namespace
from lightning.pytorch.loggers import CometLogger


def get_attr(file, attribute, key=None):
    obj = file if key is None else file[key]
    value = dict(obj.attrs).get(attribute)
    if np.issubdtype(type(value), np.integer):
        value = int(value)
    if isinstance(value, str):
        with suppress(json.decoder.JSONDecodeError, TypeError):
            value = json.loads(value)
    return value


class SaveConfigCallback(Callback):
    def __init__(
        self,
        parser: LightningArgumentParser,
        config: Namespace,
        config_filename: str = "config.yaml",
        overwrite: bool = False,
        multifile: bool = False,
    ) -> None:
        """Saves a LightningCLI config to the log_dir when training starts.

        Parameters
        ----------
        parser : LightningArgumentParser
            The parser object used to parse the configuration.
        config : Namespace
            The parsed configuration that will be saved.
        config_filename : str, optional
            Filename for the config file.
        overwrite : bool, optional
            Whether to overwrite an existing config file.
        multifile : bool, optional
            When input is multiple config files, saved config

        Raises
        ------
        RuntimeError:
            If the config file already exists in the directory
            to avoid overwriting a previous run

        """
        super().__init__()
        self.parser = parser
        self.config = config
        self.config_filename = config_filename
        self.overwrite = overwrite
        self.multifile = multifile
        self.already_saved = False

    def setup(self, trainer: Trainer, pl_module: LightningModule, stage: str) -> None:
        # save only on rank zero to avoid race conditions.
        if stage != "fit" or self.already_saved:
            return

        self.trainer = trainer
        self.plm = pl_module

        # get path info
        if "s3:/" in trainer.log_dir[:4]:
            from s3fs import S3FileSystem
            from s3path import S3Path

            self.use_S3 = True
            self.make_path = S3Path
            log_dir = self.make_path("/" + trainer.log_dir.replace("s3://", "").replace("s3:/", ""))
            self.S3_session = S3FileSystem(anon=False)
        elif "s3:/" in trainer.log_dir:
            raise ValueError(
                f"trainer.log_dir should start with 's3:/', instead of {trainer.log_dir}"
            )
        else:
            self.use_S3 = False
            self.make_path = Path
            log_dir = self.make_path(trainer.log_dir)

        assert log_dir is not None
        trainer.timestamp = log_dir.name
        # broadcast whether to fail to all ranks
        config_path = self.make_path(log_dir / self.config_filename)
        file_exists = config_path.exists()
        file_exists = trainer.strategy.broadcast(file_exists)
        if file_exists and not self.overwrite:
            raise RuntimeError(
                f"{self.__class__.__name__} expected {config_path} to NOT exist."
                " Aborting to avoid overwriting results of a previous run."
            )

        if trainer.is_global_zero:
            # save configs
            self.save_config(config_path)

            # save metadata
            self.save_metadata(config_path)

            # broadcast so that all ranks are in sync on future calls to .setup()
            self.already_saved = True

        self.already_saved = trainer.strategy.broadcast(self.already_saved)

    def save_config(self, config_path):
        # the `log_dir` needs to be created as we rely on the logger
        # to do it usually but it hasn't logged anything at this point
        config_path.parent.mkdir(parents=True, exist_ok=True)
        print("-" * 100)
        print(f"Created output dir {config_path.parent} {'on S3' if self.use_S3 else ''}")
        print("-" * 100, "\n")

        # copy the norm dict
        nd_path = self.make_path(
            config_path.parent / self.make_path(self.config.data.norm_dict).name
        )
        self.config.data.norm_dict = self.write_file(self.config.data.norm_dict, nd_path)
        # copy the class dict
        if self.config.data.class_dict:
            cd_path = self.make_path(
                config_path.parent / self.make_path(self.config.data.class_dict).name
            )
            self.config.data.class_dict = self.write_file(self.config.data.class_dict, cd_path)

        # write config
        self.write_yaml_file(self.config, config_path)

        # log files as assets
        # currently cannot save log files as assests on S3
        if isinstance(self.plm.logger, CometLogger) and not self.use_S3:
            self.plm.logger.experiment.log_asset(config_path)
            self.plm.logger.experiment.log_asset(nd_path)
            if self.config.data.class_dict:
                self.plm.logger.experiment.log_asset(cd_path)

    def save_metadata(self, config_path):
        trainer = self.trainer
        logger = self.plm.logger

        train_loader = trainer.datamodule.train_dataloader()
        val_loader = trainer.datamodule.val_dataloader()
        train_dset = train_loader.dataset
        val_dset = val_loader.dataset

        global_object = self.plm.global_object
        if input_map := self.config["data"]["input_map"]:
            global_object = input_map[global_object]

        meta = {}

        meta["train_file"] = str(train_dset.filename)
        meta["val_file"] = str(val_dset.filename)
        meta["num_train"] = len(train_dset)
        meta["num_val"] = len(val_dset)
        meta["available_train"] = len(train_dset.file[global_object])
        meta["available_val"] = len(val_dset.file[global_object])
        batch_size = train_loader.batch_size
        batch_size = batch_size if batch_size else train_loader.sampler.batch_size
        meta["batch_size"] = batch_size
        params = sum(p.numel() for p in self.plm.parameters() if p.requires_grad)
        meta["trainable_params"] = params
        meta["num_gpus"] = trainer.num_devices
        meta["gpu_ids"] = trainer.device_ids
        meta["num_workers"] = train_loader.num_workers

        with contextlib.suppress(KeyError):
            # TODO: update UPP to call attribute objects instead of jets
            # https://github.com/umami-hep/umami-preprocessing/issues/56
            meta["num_unique_jets_train"] = get_attr(train_dset.file, "unique_jets")
            meta["num_unique_jets_val"] = get_attr(val_dset.file, "unique_jets")
            meta["dsids"] = get_attr(train_dset.file, "dsids")

        meta["salt_hash"] = get_git_hash(Path(__file__).parent)
        if logger:
            meta["out_dir"] = logger.save_dir
            if not self.use_S3:
                # Currently not available on S3
                meta["log_url"] = logger.experiment.url
        if hasattr(self.trainer, "timestamp"):
            meta["timestamp"] = trainer.timestamp
        meta["torch_version"] = str(torch.__version__)
        meta["lightning_version"] = str(L.__version__)
        meta["cuda_version"] = torch.version.cuda
        meta["hostname"] = socket.gethostname()

        if logger and not self.use_S3:
            # Currently not available on S3
            logger.log_hyperparams(meta)

        # save the object classes, which is stored as an attr in the training file
        with h5py.File(meta["train_file"]) as file:
            try:
                object_classes = file[global_object].attrs["flavour_label"]
            except KeyError:
                object_classes = "not available"
            meta["object_classes"] = dict(
                zip(range(len(object_classes)), object_classes, strict=True)
            )

        with contextlib.suppress(KeyError):
            meta["jet_counts_train"] = get_attr(train_dset.file, "jet_counts")
            meta["jet_counts_val"] = get_attr(val_dset.file, "jet_counts")
            meta["pp_config_train"] = get_attr(train_dset.file, "config")
            meta["pp_config_val"] = get_attr(val_dset.file, "config")

        meta_path = self.make_path(config_path.parent / "metadata.yaml")
        self.write_dump_yaml_file(meta, meta_path)

    def write_file(self, file, store_path):
        """Writes the file on S3 at the store_path."""
        if self.use_S3:
            self.S3_session.put(file, store_path)
            outpath = file
        else:
            shutil.copyfile(file, store_path)
            outpath = str(store_path.resolve())
        return outpath

    def write_yaml_file(self, yaml_file, store_path):
        """Writes the file on S3 at the store_path."""
        if self.use_S3:
            local_path = Path("local_config.yaml")
            self.parser.save(
                yaml_file,
                str(local_path),
                skip_none=False,
                overwrite=self.overwrite,
                multifile=self.multifile,
            )
            self.S3_session.put(str(local_path), store_path)
            local_path.unlink()
        else:
            self.parser.save(
                yaml_file,
                str(store_path),
                skip_none=False,
                overwrite=self.overwrite,
                multifile=self.multifile,
            )

    def write_dump_yaml_file(self, meta, store_path):
        """Writes the file on S3 at the store_path."""
        if self.use_S3:
            local_path = Path("metadata.yaml")
            with open(str(local_path), "w") as file:
                yaml.dump(meta, file, sort_keys=False)
            self.S3_session.put(str(local_path), store_path)
            local_path.unlink()
        else:
            with open(store_path, "w") as file:
                yaml.dump(meta, file, sort_keys=False)



"""Top level training script, powered by the lightning CLI."""

import sys

from salt.main import main as main_salt
from salt.utils.muP_utils.configuration_muP import (
    generate_base_delta_config,
    store_shapes_mup,
)
from salt.utils.muP_utils.generateModel import generateModel


def main(args=None):
    """Main script for muP.
    - Should be used to instantiate needed base & delta models.
    - Can also execute the main training if global runFit is True.
    - Can also perform a check muP is working if global check is True.

    Note: recommended usage is to only use this script to instantiate
    the base & delta models. You can then run a normal salt for the main
    models, which is going to load the base & delta models.
    """
    runFit = False
    check = False
    if args is not None:
        sys.argv = [""]
        for item in args:
            if "=" in item:
                sys.argv += [item.split("=")[0]]
                sys.argv += [item.split("=")[1]]
            else:
                sys.argv += [item]

    if "--runFit" in sys.argv:
        sys.argv.remove("--runFit")
        runFit = True
    if "--check" in sys.argv:
        sys.argv.remove("--check")
        check = True
    train_file, norm_dict = None, None
    if "--data.train_file" in sys.argv:
        ind = sys.argv.index("--data.train_file")
        train_file = sys.argv[ind + 1]
    if "--data.norm_dict" in sys.argv:
        ind = sys.argv.index("--data.norm_dict")
        norm_dict = sys.argv[ind + 1]

    full_args = sys.argv[1:]
    config = sys.argv[2]
    restargs = [""]
    if len(sys.argv) > 2:
        restargs = sys.argv[3:]

    print("-" * 100)
    print("Initialising base and delta shapes")
    base_config, store_path = generate_base_delta_config(config, "base")
    delta_config, _ = generate_base_delta_config(config, "delta")

    # Execute generateModel to get base_model saved at a path in base_config
    generateModel(args=["base", "--trainer.callbacks", "", f"--config={base_config}", *restargs])
    generateModel(args=["delta", "--trainer.callbacks", "", f"--config={delta_config}", *restargs])

    # Load the models and save the shapes in filename
    store_shapes_mup(path=store_path)

    # If you want to check muP works
    if check:
        from salt.utils.muP_utils.configuration_muP import (
            check_muP,
            generate_config_muPtest,
            get_models_muPtest,
        )

        print("Instantiating muP check models")

        # Need also base and delta without wrapper for local training
        generateModel(
            args=["temp_base", "--trainer.callbacks", "", f"--config={base_config}", *restargs]
        )
        generateModel(
            args=["temp_delta", "--trainer.callbacks", "", f"--config={delta_config}", *restargs]
        )

        store_shapes_mup(path=store_path, check=check)

        # Create the variations
        variations = [32, 64, 128, 256, 512, 1024, 2056]
        configs = generate_config_muPtest(config, variations)
        configs.extend(generate_config_muPtest(config, variations, muP=False))
        for enum, mod_config in enumerate(configs):
            temp_add = "muP" if "muP" in str(mod_config.name) else "sP"

            generateModel(
                args=[
                    f"temp_{temp_add}_{enum % len(variations)}",
                    "--trainer.callbacks",
                    "",
                    f"--config={mod_config}",
                    *restargs,
                ]
            )

        if runFit:
            print("-" * 100)
            print("Running muP checks")
            assert train_file is not None, "You must give a valid train_file path"
            assert norm_dict is not None, "You must give a valid norm_dict path"

            # Retrieve the muP and SP models
            models_muP, models_sP = get_models_muPtest(variations)

            # Perform checks:
            #   - Warning, will train the models!
            #       Recommended to use a script to separately do the training
            #   - requires to manually pass the train_file and norm_dict.
            #       The "variables" dictionary is set in
            check_muP(
                models_muP,
                models_sP,
                num_train=2000,
                nsteps_training=1500,
                nsteps_coocheck=3,
                nseeds=5,
                batch_size=1000,
                doCoord=True,
                doTraining=True,
                train_file=train_file,
                norm_dict=norm_dict,
            )
        return

    # Normal execution of salt for the real model, called model.
    # The lightning_module should be able to collect filename and do:
    #       set_base_shapes(model, base_model, delta=delta_model)
    if runFit:
        print("-" * 100)
        print("Starting main training")
        main_salt(["fit", *full_args])

    # To cleam the muP produced files
    # clean_environment()


if __name__ == "__main__":
    main()
